# -*- coding: utf-8 -*-
from structuregraph_helpers.create import get_structure_graph
from structuregraph_helpers.subgraph import get_subgraphs_as_molecules

from mofdscribe.featurizers.hostguest.utils import remove_guests_from_structure


def test_remove_guests_from_structure(floating_structure):
    """Test the remove_guests_from_structure function."""
    sg = get_structure_graph(floating_structure)

    _, _, idx, _, _ = get_subgraphs_as_molecules(sg)
    structure = remove_guests_from_structure(floating_structure, idx)
    assert len(structure) == len(floating_structure) - 5



from .bu import LSOP  # noqa: F401
from .bu import PMI1  # noqa: F401
from .bu import PMI2  # noqa: F401
from .bu import PMI3  # noqa: F401
from .bu import Asphericity  # noqa: F401
from .bu import BUFeaturizer  # noqa: F401
from .bu import BUMatch  # noqa: F401
from .bu import CompositionStats  # noqa: F401
from .bu import DiskLikeness  # noqa: F401
from .bu import Eccentricity  # noqa: F401
from .bu import InertialShapeFactor  # noqa: F401
from .bu import MOFBBs  # noqa: F401
from .bu import NConf20  # noqa: F401
from .bu import PairwiseDistanceHist  # noqa: F401
from .bu import RadiusOfGyration  # noqa: F401
from .bu import RDKitAdaptor  # noqa: F401
from .bu import RodLikeness  # noqa: F401
from .bu import SphereLikeness  # noqa: F401
from .bu import SpherocityIndex  # noqa: F401
from .chemistry import AMD  # noqa: F401
from .chemistry import APRDF  # noqa: F401
from .chemistry import RACS  # noqa: F401
from .chemistry import EnergyGridHistogram  # noqa: F401
from .chemistry import Henry  # noqa: F401
from .chemistry import PartialChargeHistogram  # noqa: F401
from .chemistry import PartialChargeStats  # noqa: F401
from .chemistry import PriceLowerBound  # noqa: F401
from .pore import AccessibleVolume  # noqa: F401
from .pore import PoreDiameters  # noqa: F401
from .pore import PoreSizeDistribution  # noqa: F401
from .pore import RayTracingHistogram  # noqa: F401
from .pore import SurfaceArea  # noqa: F401
from .topology import AtomCenteredPH, PHHist, PHImage, PHStats, PHVect  # noqa: F401



# -*- coding: utf-8 -*-
"""Test ARCMOF dataset."""
import numpy as np
from pymatgen.core import IStructure

from mofdscribe.datasets.arcmof_dataset import ARCMOFDataset


def test_arcmof():
    """Ensure we can instantiate the ARCMOF dataset and access a few key methods."""
    arc = ARCMOFDataset()
    assert isinstance(list(arc.get_structures([1]))[0], IStructure)
    assert isinstance(arc.get_labels([1]), np.ndarray)

    # make sure we can get a subset of the dataset
    arc_subset = arc.get_subset([1, 2, 3, 8])
    assert isinstance(arc_subset, ARCMOFDataset)
    assert len(arc_subset._df) == 4
    assert (
        arc._df.iloc[[1, 2, 3, 8]]["info.name"].values == arc_subset._df["info.name"].values
    ).all()
    assert list(arc.get_structures([1]))[0] == list(arc_subset.get_structures([0]))[0]



# -*- coding: utf-8 -*-
"""Basic raspa output parser.

code from https://github.com/lsmo-epfl/aiida-raspa/blob/develop/aiida_raspa/utils/base_parser.py

MIT License
Copyright 2018-2022, LABORATORY OF MOLECULAR SIMULATION (LSMO),
ECOLE POLYTECHNIQUE FEDERALE DE LAUSANNE
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:
The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.
THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""
import re
from math import isinf, isnan

float_base = float


def float(number):
    number = float_base(number)
    return number if not any((isnan(number), isinf(number))) else None


KELVIN_TO_KJ_PER_MOL = float(8.314464919 / 1000.0)  # exactly the same as Raspa

# manage block of the first type
# --------------------------------------------------------------------------------------------
BLOCK_1_LIST = [
    ("Average Volume:", "cell_volume", (1, 2, 4), 0),
    ("Average Density:", "adsorbate_density", (1, 2, 4), 0),
    # ("Average Heat Capacity", "framework_heat_capacity", (1, 2, 4), 0), # misleading property!
    ("Enthalpy of adsorption:", "enthalpy_of_adsorption", (1, 4, 3), 4),
    ("Tail-correction energy:", "tail_correction_energy", (1, 2, 4), 0)
    # ("Total energy:$", "total_energy", (1, 2, 4), 0), # not important property!
]

# block of box properties.
BOX_PROP_LIST = [
    ("Average Box-lengths:", "box"),
]


def parse_block1(flines, result_dict, prop, value=1, unit=2, dev=4):
    """Parse block.

    Parses blocks that look as follows::

        Average Volume:
        =================
            Block[ 0]        12025.61229 [A^3]
            Block[ 1]        12025.61229 [A^3]
            Block[ 2]        12025.61229 [A^3]
            Block[ 3]        12025.61229 [A^3]
            Block[ 4]        12025.61229 [A^3]
            ------------------------------------------------------------------------------
            Average          12025.61229 [A^3] +/-            0.00000 [A^3]

    Args:
        flines (list): list of lines to parse
        result_dict (dict): dictionary to store results
        prop (str): property to store
        value (int): index of value in line
        unit (int): index of unit in line
        dev (int): index of deviation in line
    """
    for line in flines:
        if "Average" in line:
            result_dict[prop + "_average"] = float(line.split()[value])
            result_dict[prop + "_unit"] = re.sub(r"[{}()\[\]]", "", line.split()[unit])
            result_dict[prop + "_dev"] = float(line.split()[dev])
            break


# manage energy reading
# --------------------------------------------------------------------------------------------
ENERGY_CURRENT_LIST = [
    ("Host/Adsorbate energy:", "host/ads", "tot"),
    ("Host/Adsorbate VDW energy:", "host/ads", "vdw"),
    ("Host/Adsorbate Coulomb energy:", "host/ads", "coulomb"),
    ("Adsorbate/Adsorbate energy:", "ads/ads", "tot"),
    ("Adsorbate/Adsorbate VDW energy:", "ads/ads", "vdw"),
    ("Adsorbate/Adsorbate Coulomb energy:", "ads/ads", "coulomb"),
]

ENERGY_AVERAGE_LIST = [
    ("Average Adsorbate-Adsorbate energy:", "ads/ads"),
    ("Average Host-Adsorbate energy:", "host/ads"),
]


def parse_block_energy(flines, res_dict, prop):
    """Parse energy block.

    Parse block that looks as follows::

        Average Adsorbate-Adsorbate energy:
        ===================================
            Block[ 0] -443.23204         Van der Waals: -443.23204         Coulomb: 0.00000            [K]
            Block[ 1] -588.20205         Van der Waals: -588.20205         Coulomb: 0.00000            [K]
            Block[ 2] -538.43355         Van der Waals: -538.43355         Coulomb: 0.00000            [K]
            Block[ 3] -530.00960         Van der Waals: -530.00960         Coulomb: 0.00000            [K]
            Block[ 4] -484.15106         Van der Waals: -484.15106         Coulomb: 0.00000            [K]
            ------------------------------------------------------------------------------
            Average   -516.80566         Van der Waals: -516.805659        Coulomb: 0.00000            [K]
                  +/- 98.86943                      +/- 98.869430               +/- 0.00000            [K]

    Args:
        flines (list): list of lines to parse
        res_dict (dict): dictionary to store results
        prop (str): property to store
    """
    for line in flines:
        if "Average" in line:
            res_dict["energy_{}_tot_average".format(prop)] = (
                float(line.split()[1]) * KELVIN_TO_KJ_PER_MOL
            )
            res_dict["energy_{}_vdw_average".format(prop)] = (
                float(line.split()[5]) * KELVIN_TO_KJ_PER_MOL
            )
            res_dict["energy_{}_coulomb_average".format(prop)] = (
                float(line.split()[7]) * KELVIN_TO_KJ_PER_MOL
            )
        if "+/-" in line:
            res_dict["energy_{}_tot_dev".format(prop)] = (
                float(line.split()[1]) * KELVIN_TO_KJ_PER_MOL
            )
            res_dict["energy_{}_vdw_dev".format(prop)] = (
                float(line.split()[3]) * KELVIN_TO_KJ_PER_MOL
            )
            res_dict["energy_{}_coulomb_dev".format(prop)] = (
                float(line.split()[5]) * KELVIN_TO_KJ_PER_MOL
            )
            return


# manage lines with components
LINES_WITH_COMPONENT_LIST = [
    (" Average Widom Rosenbluth-weight:", "widom_rosenbluth_factor"),
    (" Average chemical potential: ", "chemical_potential"),
    (" Average Henry coefficient: ", "henry_coefficient"),
    (" Average  <U_gh>_1-<U_h>_0:", "adsorption_energy_widom"),
]


def parse_lines_with_component(res_components, components, line, prop):
    """Parse lines that contain components"""
    for i, component in enumerate(components):
        if "[" + component + "]" in line:
            words = line.split()
            res_components[i][prop + "_unit"] = re.sub(r"[{}()\[\]]", "", words[-1])
            res_components[i][prop + "_dev"] = float(words[-2])
            res_components[i][prop + "_average"] = float(words[-4])


def parse_base_output(output_abs_path, system_name, ncomponents):  # noqa: C901
    """Parse RASPA output file: it is divided in different parts, whose start/end is carefully documented."""
    warnings = []
    res_per_component = []
    for _ in range(ncomponents):
        res_per_component.append({})
    result_dict = {"exceeded_walltime": False}

    with open(output_abs_path, "r") as fobj:

        # 1st parsing part: input settings
        # --------------------------------
        # from: start of file
        # to: "Current (initial full energy) Energy Status"

        icomponent = 0
        component_names = []
        res_cmp = res_per_component[0]
        for line in fobj:
            if "Component" in line and "molecule)" in line:
                component_names.append(line.split()[2][1:-1])
                if "(Adsorbate" in line:
                    res_cmp["molecule_type"] = "adsorbate"
                elif "(Cation" in line:
                    res_cmp["molecule_type"] = "cation"

            # Consider to change it with parse_line()
            if "Conversion factor molecules/unit cell -> mol/kg:" in line:
                res_cmp["conversion_factor_molec_uc_to_mol_kg"] = float(line.split()[6])
                res_cmp["conversion_factor_molec_uc_to_mol_kg_unit"] = "(mol/kg)/(molec/uc)"
            # this line was corrected in Raspa's commit c1ad4de (Nov19), since "gr/gr" should read "mg/g"
            if (
                "Conversion factor molecules/unit cell -> gr/gr:" in line
                or "Conversion factor molecules/unit cell -> mg/g:" in line
            ):
                res_cmp["conversion_factor_molec_uc_to_mg_g"] = float(line.split()[6])
                res_cmp["conversion_factor_molec_uc_to_mg_g_unit"] = "(mg/g)/(molec/uc)"
            if "Conversion factor molecules/unit cell -> cm^3 STP/gr:" in line:
                res_cmp["conversion_factor_molec_uc_to_cm3stp_gr"] = float(line.split()[7])
                res_cmp["conversion_factor_molec_uc_to_cm3stp_gr_unit"] = "(cm^3_STP/gr)/(molec/uc)"
            if "Conversion factor molecules/unit cell -> cm^3 STP/cm^3:" in line:
                res_cmp["conversion_factor_molec_uc_to_cm3stp_cm3"] = float(line.split()[7])
                res_cmp[
                    "conversion_factor_molec_uc_to_cm3stp_cm3_unit"
                ] = "(cm^3_STP/cm^3)/(molec/uc)"
            if "MolFraction:" in line:
                res_cmp["mol_fraction"] = float(line.split()[1])
                res_cmp["mol_fraction_unit"] = "-"
            if "Partial pressure:" in line:
                res_cmp["partial_pressure"] = float(line.split()[2])
                res_cmp["partial_pressure_unit"] = "Pa"
            if "Partial fugacity:" in line:
                res_cmp["partial_fugacity"] = float(line.split()[2])
                res_cmp["partial_fugacity_unit"] = "Pa"
                icomponent += 1
                if icomponent < ncomponents:
                    res_cmp = res_per_component[icomponent]
            if "Framework Density" in line:
                result_dict["framework_density"] = line.split()[2]
                result_dict["framework_density_unit"] = re.sub(r"[{}()\[\]]", "", line.split()[3])
            if "Current (initial full energy) Energy Status" in line:
                break

        # 2nd parsing part: initial and final configurations
        # --------------------------------------------------
        # from: "Current (initial full energy) Energy Status"
        # to: "Average properties of the system"

        reading = "initial"
        result_dict["energy_unit"] = "kJ/mol"

        for line in fobj:
            # Understand if it is the initial or final "Current Energy Status" section
            if "Current (full final energy) Energy Status" in line:
                reading = "final"

            # Read the entries of "Current Energy Status" section
            if reading:
                for parse in ENERGY_CURRENT_LIST:
                    if parse[0] in line:
                        result_dict["energy_{}_{}_{}".format(parse[1], parse[2], reading)] = (
                            float(line.split()[-1]) * KELVIN_TO_KJ_PER_MOL
                        )
                        if parse[1] == "ads/ads" and parse[2] == "coulomb":
                            reading = None

            if "Average properties of the system" in line:
                break

        # 3rd parsing part: average system properties
        # --------------------------------------------------
        # from: "Average properties of the system"
        # to: "Number of molecules"

        for line in fobj:
            for parse in BLOCK_1_LIST:
                if parse[0] in line:
                    parse_block1(fobj, result_dict, parse[1], *parse[2])
                    # I assume here that properties per component are present furhter in the output file.
                    # so I need to skip some lines:
                    skip_nlines_after = parse[3]
                    while skip_nlines_after > 0:
                        line = next(fobj)
                        skip_nlines_after -= 1
                    for i, cmpnt in enumerate(component_names):
                        # The order of properties per molecule is the same as the order of molecules in the
                        # input file. So if component name was not found in the next line, I break the loop
                        # immidiately as there is no reason to continue it
                        line = next(fobj)
                        if cmpnt in line:
                            parse_block1(fobj, res_per_component[i], parse[1], *parse[2])
                        else:
                            break
                        skip_nlines_after = parse[3]
                        while skip_nlines_after > 0:
                            line = next(fobj)
                            skip_nlines_after -= 1

                    continue  # no need to perform further checks, propperty has been found already
            for parse in ENERGY_AVERAGE_LIST:
                if parse[0] in line:
                    parse_block_energy(fobj, result_dict, prop=parse[1])
                    continue  # no need to perform further checks, propperty has been found already
            for parse in BOX_PROP_LIST:
                if parse[0] in line:
                    # parse three cell vectors
                    parse_block1(fobj, result_dict, prop="box_ax", value=2, unit=3, dev=5)
                    parse_block1(fobj, result_dict, prop="box_by", value=2, unit=3, dev=5)
                    parse_block1(fobj, result_dict, prop="box_cz", value=2, unit=3, dev=5)
                    # parsee angles between the cell vectors
                    parse_block1(fobj, result_dict, prop="box_alpha", value=3, unit=4, dev=6)
                    parse_block1(fobj, result_dict, prop="box_beta", value=3, unit=4, dev=6)
                    parse_block1(fobj, result_dict, prop="box_gamma", value=3, unit=4, dev=6)

            if "Number of molecules:" in line:
                break

        # 4th parsing part: average molecule properties
        # --------------------------------------------------
        # from: "Number of molecules"
        # to: end of file

        icomponent = 0
        for line in fobj:
            # Consider to change it with parse_line?
            if "Average loading absolute [molecules/unit cell]" in line:
                res_per_component[icomponent]["loading_absolute_average"] = float(line.split()[5])
                res_per_component[icomponent]["loading_absolute_dev"] = float(line.split()[7])
                res_per_component[icomponent]["loading_absolute_unit"] = "molecules/unit cell"
            elif "Average loading excess [molecules/unit cell]" in line:
                res_per_component[icomponent]["loading_excess_average"] = float(line.split()[5])
                res_per_component[icomponent]["loading_excess_dev"] = float(line.split()[7])
                res_per_component[icomponent]["loading_excess_unit"] = "molecules/unit cell"
                icomponent += 1
            if icomponent >= ncomponents:
                break

        for line in fobj:
            for to_parse in LINES_WITH_COMPONENT_LIST:
                if to_parse[0] in line:
                    parse_lines_with_component(
                        res_per_component, component_names, line, to_parse[1]
                    )

        # Assigning to None all the quantities that are meaningless if not running a Widom insertion calculation
        for res_comp in res_per_component:
            for prop in ["henry_coefficient", "widom_rosenbluth_factor", "chemical_potential"]:
                if res_comp["{}_dev".format(prop)] == 0.0:
                    res_comp["{}_average".format(prop)] = None
                    res_comp["{}_dev".format(prop)] = None

            # The section "Adsorption energy from Widom-insertion" is not showing in the output if no widom is performed
            if "adsorption_energy_widom_average" not in res_comp:
                res_comp["adsorption_energy_widom_unit"] = "kJ/mol"
                res_comp["adsorption_energy_widom_dev"] = None
                res_comp["adsorption_energy_widom_average"] = None

    return_dictionary = {"general": result_dict, "components": {}}

    for name, value in zip(component_names, res_per_component):
        return_dictionary["components"][name] = value

    # Parsing all the warning that are printed in the output file, avoiding redoundancy
    with open(output_abs_path, "r") as fobj:
        for line in fobj:
            if "WARNING" in line:
                warning_touple = (system_name, line)
                if warning_touple not in warnings:
                    warnings.append(warning_touple)
    return return_dictionary, warnings



"""Featurize a molecule using SMARTS matches."""
from functools import partial
from typing import Collection, List, Optional

from rdkit import Chem

from mofdscribe.featurizers.bu.rdkitadaptor import RDKitAdaptor


def number_smart_matches(mol, smarts: Collection[str]) -> int:
    """Count the number of SMARTS matches in a molecule.

    This can be useful if we have some prior knowledge
    about which substructures might be interesting/relevant.

    Args:
        mol (rdkit.Chem.rdchem.Mol): RDKit molecule.
        smarts (Collection[str]): SMARTS patterns to match.

    Returns:
        int: Number of SMARTS matches.
    """
    s = ",".join("$(" + s + ")" for s in smarts)
    smarts_mol = Chem.MolFromSmarts("[" + s + "]")
    return len(mol.GetSubstructMatches(smarts_mol))


class SmartsMatchCounter(RDKitAdaptor):
    """Count the number of SMARTS matches in a molecule.

    This can be useful if we have some prior knowledge
    about which substructures might be interesting/relevant.
    For instance, you might want to count the number of
    carboxylic acid groups in a molecule.
    """

    # ToDo: Perhaps normalize by the length of the SMARTS substructure
    def __init__(self, smarts: Collection[str], feature_labels: Optional[Collection[str]]) -> None:
        """Construct a new SmartsMatchCounter.

        Args:
            smarts (Collection[str]): SMARTS patterns to match.
            feature_labels (str, optional): Feature labels.
                If None, the SMARTS patterns are concatenated to a labels.
        """
        featurizer = partial(number_smart_matches, smarts=smarts)
        if feature_labels is None:
            smarts_string = "_".join(smarts)
            feature_labels = [f"smarts_{smarts_string}"]
        super().__init__(featurizer, feature_labels)


class AcidGroupCounter(SmartsMatchCounter):
    """Count the number of acidic groups in a molecule.

    SMARTS patterns are taken from the Mordred package.
    """

    def __init__(self) -> None:
        """Construct a new AcidGroupCounter."""
        smarts = ["[O;H1]-[C,S,P]=O", "[*;-;!$(*~[*;+])]", "[NH](S(=O)=O)C(F)(F)F", "n1nnnc1"]
        super().__init__(smarts, feature_labels=["acid_groups"])

    def citations(self) -> List[str]:
        return super().citations() + [
            "@article{Moriwaki_2018,"
            "doi = {10.1186/s13321-018-0258-y},"
            "url = {https://doi.org/10.1186%2Fs13321-018-0258-y},"
            "year = 2018,"
            "month = {feb},"
            "publisher = {Springer Science and Business Media {LLC}},"
            "volume = {10},"
            "number = {1},"
            "author = {Hirotomo Moriwaki and Yu-Shi Tian and Norihito Kawashita and Tatsuya Takagi},"
            "title = {Mordred: a molecular descriptor calculator},"
            "journal = {J Cheminform}"
            "}"
        ]

    def implementors(self) -> List[str]:
        return super().implementors() + ["Moriwaki H, Tian Y-S, Kawashita N, Takagi T"]


class BaseGroupCounter(SmartsMatchCounter):
    """Count the number of basic groups in a molecule.

    SMARTS pattern taken from the Mordred package
    """

    def __init__(self) -> None:
        """Construct a new BaseGroupCounter."""
        smarts = [
            "[NH2]-[CX4]",
            "[NH](-[CX4])-[CX4]",
            "N(-[CX4])(-[CX4])-[CX4]",
            "[*;+;!$(*~[*;-])]",
            "N=C-N",
            "N-C=N",
        ]
        super().__init__(smarts, feature_labels=["base_groups"])

    def citations(self) -> List[str]:
        return super().citations() + [
            "@article{Moriwaki_2018,"
            "doi = {10.1186/s13321-018-0258-y},"
            "url = {https://doi.org/10.1186%2Fs13321-018-0258-y},"
            "year = 2018,"
            "month = {feb},"
            "publisher = {Springer Science and Business Media {LLC}},"
            "volume = {10},"
            "number = {1},"
            "author = {Hirotomo Moriwaki and Yu-Shi Tian and Norihito Kawashita and Tatsuya Takagi},"
            "title = {Mordred: a molecular descriptor calculator},"
            "journal = {J Cheminform}"
            "}"
        ]

    def implementors(self) -> List[str]:
        return super().implementors() + ["Moriwaki H, Tian Y-S, Kawashita N, Takagi T"]



from mofdscribe.featurizers.hostguest.guest_aprdf import GuestCenteredAPRDF


def test_guest_centered_aprdf(floating_structure):
    """Test the GuestCenteredAPRDF."""
    featurizer = GuestCenteredAPRDF()

    features = featurizer.featurize(floating_structure)
    assert len(features) == (20 - 2) / 0.25 * 2 * 3
    assert len(featurizer.feature_labels()) == (20 - 2) / 0.25 * 2 * 3



# -*- coding: utf-8 -*-
"""Test the basic benchmark scaffolding."""
import os
from typing import Dict, Optional

import numpy as np
import pytest
from pymatgen.core import Structure
from sklearn.linear_model import LinearRegression
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler

from mofdscribe.bench.mofbench import BenchResult, MOFBench, MOFBenchRegression, id_for_bench_result
from mofdscribe.datasets import CoREDataset
from mofdscribe.splitters import ClusterStratifiedSplitter
from mofdscribe.splitters.splitters import BaseSplitter


@pytest.mark.xdist_group(name="core-ds")
class MyDummyModel:
    """Dummy model."""

    def __init__(self, lr_kwargs: Optional[Dict] = None):
        """Initialize the model.

        Args:
            lr_kwargs (Optional[Dict], optional): Keyword arguments
                that are passed to the linear regressor.
                Defaults to None.
        """
        self.model = Pipeline(
            [("scaler", StandardScaler()), ("lr", LinearRegression(**(lr_kwargs or {})))]
        )

    def featurize(self, s: Structure):
        """You might want to use a lookup in some dataframe instead.

        Or use some mofdscribe featurizers.
        """
        return s.density

    def fit(self, idx, structures, y):
        self.log({"info": "hello"})
        x = np.array([self.featurize(s) for s in structures]).reshape(-1, 1)
        self.model.fit(x, y)

    def predict(self, idx, structures):
        x = np.array([self.featurize(s) for s in structures]).reshape(-1, 1)
        return self.model.predict(x)


@pytest.mark.xdist_group(name="core-ds")
def test_mofbench(tmp_path_factory):
    """Test the MOFBench class."""
    ds = CoREDataset()
    with pytest.raises(TypeError):
        bench = MOFBench(
            model=MyDummyModel(),
            ds=ds,
            splitter=ClusterStratifiedSplitter(
                CoREDataset(), feature_names=list(ds.available_features)
            ),
            target=["outputs.logKH_CO2"],
        )
    bench = MOFBenchRegression(
        model=MyDummyModel(),
        ds=ds,
        name="my model",
        task="logKH_CO2_id",
        splitter=BaseSplitter(CoREDataset(), sample_frac=0.001),
        target=["outputs.logKH_CO2"],
        debug=True,
        k=2,
    )
    assert isinstance(bench, MOFBench)

    report = bench.bench()
    assert isinstance(report, BenchResult)
    assert len(report.logs) == 2
    assert report.logs[0] == {"info": "hello"}
    assert isinstance(report.json(), str)
    path = os.path.join(tmp_path_factory.mktemp("report"))
    report.save_json(path)
    read_bench = BenchResult.parse_file(
        os.path.join(path, f"{id_for_bench_result(report)}.json"),
    )
    assert isinstance(read_bench, BenchResult)



# -*- coding: utf-8 -*-
"""Test partial charge featurizer."""
from mofdscribe.featurizers.chemistry.partialchargehistogram import PartialChargeHistogram

from ..helpers import is_jsonable


def test_partialchargehistogram(hkust_structure, irmof_structure):
    """Make sure that the featurization works for typical MOFs
    and the number of features is as expected.
    """
    pch = PartialChargeHistogram()
    assert len(pch.feature_labels()) == 16

    for structure in [hkust_structure, irmof_structure]:
        features = pch.featurize(structure)
        assert len(features) == 16
    assert is_jsonable(dict(zip(pch.feature_labels(), features)))
    assert features.ndim == 1



# -*- coding: utf-8 -*-
"""Subset of the QMOF dataset."""
import os
from typing import Collection, Optional, Tuple

import numpy as np
import pandas as pd
from loguru import logger

from mofdscribe.constants import MOFDSCRIBE_PYSTOW_MODULE
from mofdscribe.datasets.checks import check_all_file_exists, length_check
from mofdscribe.datasets.dataset import AbstractStructureDataset
from mofdscribe.datasets.utils import compress_dataset

__all__ = ["QMOFDataset"]


class QMOFDataset(AbstractStructureDataset):
    """
    Exposes the QMOF dataset by Rosen et al. [Rosen2021]_ [Rosen2022]_ .

    Currently based on v14 of the QMOF dataset.

    To reduce the risk of data leakage, we (by default) also only keep one representative
    structure for a "base refcode" (i.e. the first five letters of a refcode).
    For instance, the base refcode for IGAHED001 is IGAHED. Structures with same
    base refcode but different refcodes are often different refinements, or measurements
    at different temperatures and hence chemically quite similar.
    For instance, in the QMOF dataset `the basecode BOJKAM appears
    four times
    <https://materialsproject.org/mofs?_sort=data.lcd.value&data__csdRefcode__contains=BOJKAM>`_.
    Additionally, we (by default) only keep one structure per "structure hash" which
    is an approximate graph-isomoprhism check, assuming the VESTA bond thresholds
    for the derivation of the structure graph.

    Note that Rosen et al. already performed some deduplication using the pymatgen `StructureMatcher`.
    Our de-duplication is a bit more aggressive, and might be too aggressive in some cases.

    .. warning::
        Even though we performed some basic sanity checks and Rosen et al.
        included checks to ensure high-fidelity structures, there might
        still be some structures that are not chemically reasonable.
        Also, even though we only keep one structure per base refcode, there is still
        potential for data leakge. We urge users to still drop duplicates (or close neighbors)
        after featurization.

    This dataset is available in different flavors:

    * ``"all"``: the full dataset, all original QMOF structures for which we
        could compute features and hashes
    * ``"csd"``:  the subset which comes from the CSD and for which
        we could retrieve publication years.
    * ``"gcmc"``: the subset for which we performed grand canonical Monte Carlo
        simulations
    * ``"gcmc-csd"``: the subset for which we performed grand canonical Monte Carlo
        simulations and for which we could retrieve publication years.

    Currently, we expose the following labels:

        * outputs.pbe.energy_total
        * outputs.pbe.energy_vdw
        * outputs.pbe.energy_elec
        * outputs.pbe.net_magmom
        * outputs.pbe.bandgap
        * outputs.pbe.cbm
        * outputs.pbe.vbm
        * outputs.pbe.directgap
        * outputs.pbe.bandgap_spins
        * outputs.pbe.cbm_spins
        * outputs.pbe.vbm_spins
        * outputs.pbe.directgap_spins
        * outputs.hle17.energy_total
        * outputs.hle17.energy_vdw
        * outputs.hle17.energy_elec
        * outputs.hle17.net_magmom
        * outputs.hle17.bandgap
        * outputs.hle17.cbm
        * outputs.hle17.vbm
        * outputs.hle17.directgap
        * outputs.hle17.bandgap_spins
        * outputs.hle17.cbm_spins
        * outputs.hle17.vbm_spins
        * outputs.hle17.directgap_spins
        * outputs.hse06_10hf.energy_total
        * outputs.hse06_10hf.energy_vdw
        * outputs.hse06_10hf.energy_elec
        * outputs.hse06_10hf.net_magmom
        * outputs.hse06_10hf.bandgap
        * outputs.hse06_10hf.cbm
        * outputs.hse06_10hf.vbm
        * outputs.hse06_10hf.directgap
        * outputs.hse06_10hf.bandgap_spins
        * outputs.hse06_10hf.cbm_spins
        * outputs.hse06_10hf.vbm_spins
        * outputs.hse06_10hf.directgap_spins
        * outputs.hse06.energy_total
        * outputs.hse06.energy_vdw
        * outputs.hse06.energy_elec
        * outputs.hse06.net_magmom
        * outputs.hse06.bandgap
        * outputs.hse06.cbm
        * outputs.hse06.vbm
        * outputs.hse06.directgap
        * outputs.hse06.bandgap_spins
        * outputs.hse06.cbm_spins
        * outputs.hse06.vbm_spins
        * outputs.hse06.directgap_spins
        * outputs.CO2_Henry_coefficient
        * outputs.CO2_adsorption_energy
        * outputs.N2_Henry_coefficient
        * outputs.N2_adsorption_energy
        * outputs.CO2_parasitic_energy_(coal)
        * outputs.Gravimetric_working_capacity_(coal)
        * outputs.Volumetric_working_capacity_(coal)
        * outputs.CO2_parasitic_energy_(nat_gas)
        * outputs.Gravimetric_working_capacity_(nat_gas)
        * outputs.Volumetric_working_capacity_(nat_gas)
        * outputs.Final_CO2_purity_(nat_gas)
        * outputs.CH4_Henry_coefficient
        * outputs.CH4_adsorption_energy
        * outputs.Enthalphy_of_Adsorption__at__58_bar,_298K
        * outputs.Enthalphy_of_Adsorption__at__65bar--298K
        * outputs.Working_capacity_vol_(58-65bar--298K)
        * outputs.Working_capacity_mol_(58-65bar--298K)
        * outputs.Working_capacity_fract_(58-65bar--298K)
        * outputs.Working_capacity_wt%_(58-65bar--298K)
        * outputs.O2_Henry_coefficient
        * outputs.O2_adsorption_energy
        * outputs.Enthalphy_of_Adsorption__at__5_bar,_298K
        * outputs.Enthalphy_of_Adsorption__at__140bar--298K
        * outputs.Working_capacity_vol_(5-140bar--298K)
        * outputs.Working_capacity_mol_(5-140bar--298K)
        * outputs.Working_capacity_fract_(5-140bar--298K)
        * outputs.Working_capacity_wt%_(5-140bar--298K)
        * outputs.Xe_Henry_coefficient
        * outputs.Xe_adsorption_energy
        * outputs.Kr_Henry_coefficient
        * outputs.Kr_adsorption_energy
        * outputs.Xe--Kr_selectivity__at__298K
        * outputs.Working_capacity_g--L_(5-100bar--298-198K)
        * outputs.Working_capacity_g--L_(5-100bar--77K)
        * outputs.Working_capacity_g--L_(1-100bar--77K)
        * outputs.Working_capacity_wt%_(5-100bar--298-198K)
        * outputs.Working_capacity_wt%_(5-100bar--77K)
        * outputs.Working_capacity_wt%_(1-100bar--77K)
        * outputs.H2S_Henry_coefficient
        * outputs.H2S_adsorption_energy
        * outputs.H2O_Henry_coefficient
        * outputs.H2O_adsorption_energy
        * outputs.H2S--H2O_selectivity__at__298K
        * outputs.CH4--N2_selectivity__at__298K

    Note that many of the gas adsorption data are :py:obj:`numpy.nan` because the pores
    are not accessible to the guest molecules. Depending on your application you might want
    to fill them with zeros or drop them.

    .. warning::

        The class will load almost 1GB of data into memory.

    .. warning::

        By default, the values will be sorted by the PBE total energy

    References:
        .. [Rosen2021] `Rosen, A. S.; Iyer, S. M.; Ray, D.; Yao, Z.; Aspuru-Guzik, A.; Gagliardi, L.;
            Notestein, J. M.; Snurr, R. Q. Machine Learning the Quantum-Chemical Properties
            of Metal–Organic Frameworks for Accelerated Materials Discovery.
            Matter 2021, 4 (5), 1578–1597. <https://doi.org/10.1016/j.matt.2021.02.015>`_

        .. [Rosen2022] `Rosen, A. S.; Fung, V.; Huck, P.; O'Donnell, C. T.; Horton, M. K.; Truhlar, D. G.;
            Persson, K. A.; Notestein, J. M.; Snurr, R. Q.
            High-Throughput Predictions of Metal–Organic Framework Electronic Properties:
            Theoretical Challenges, Graph Neural Networks, and Data Exploration.
            npj Computational Materials, 8, 112.
            <https://doi.org/10.1038/s41524-022-00796-6>`_

    """

    # we expect this len for the full dataset
    _files = {
        "v0.0.1": {
            "df": "https://zenodo.org/record/7031397/files/data.json?download=1",
            "structures": "https://zenodo.org/record/7031397/files/structures.tar.gz?download=1",
            "expected_length": 15042,
            "flavors": {
                "all": 15042,
                "csd": 6311,
                "gcmc": 5321,
                "csd-gcmc": 2295,
            },
        }
    }

    def __init__(
        self,
        version: str = "v0.0.1",
        flavor: str = "all",
        drop_basename_duplicates: bool = True,
        drop_graph_duplicates: bool = True,
        subset: Optional[Collection[int]] = None,
        drop_nan: bool = False,
    ):
        """Construct an instance of the QMOF dataset.

        Args:
            version (str): version number to use.
                Defaults to "v0.0.1".
            flavor (str): flavor of the dataset to use.
                Accepted values are "all", "csd", "gcmc", and "csd-gcmc".
                Defaults to "all".
            drop_basename_duplicates (bool): If True, keep only one structure
                per CSD basename. Defaults to True.
            drop_graph_duplicates (bool): If True, keep only one structure
                per decorated graph hash. Defaults to True.
            subset (Optional[Collection[int]]): indices of the structures to include.
                This is useful for subsampling the dataset. Defaults to None.
            drop_nan (bool): If True, drop rows with NaN values in features or hashes.
                Defaults to False.

        Raises:
            ValueError: If the provided version number is not available.
        """
        self._drop_basename_duplicates = drop_basename_duplicates
        self._drop_nan = drop_nan
        self._drop_graph_duplicates = drop_graph_duplicates
        self._flavor = flavor
        if version not in self._files:
            raise ValueError(
                f"Version {version} not available. Available versions: {list(self._files.keys())}"
            )
        if flavor not in self._files[version]["flavors"]:
            raise ValueError(
                f"Flavor {flavor} not available. Available flavors: {list(self._files[version]['flavors'].keys())}"
            )
        self.version = version

        # download the data for the largest set ("all")
        self._structure_dir = MOFDSCRIBE_PYSTOW_MODULE.ensure_untar(
            "QMOF",
            self.version,
            name="structures.tar.gz",
            url=self._files[version]["structures"],
        )

        self._df = pd.DataFrame(
            MOFDSCRIBE_PYSTOW_MODULE.ensure_json(
                "QMOF", self.version, name="data.json", url=self._files[version]["df"]
            )
        ).reset_index(drop=True)
        compress_dataset(self._df)
        length_check(self._df, self._files[version]["expected_length"])

        # we sort by the PBE energy to make sure we keep always the lowest in energy
        self._df = self._df.sort_values(by="outputs.pbe.energy_total")
        if drop_basename_duplicates:
            old_len = len(self._df)
            self._df = self._df.drop_duplicates(subset=["info.basename"], keep="first")
            logger.debug(
                f"Dropped {old_len - len(self._df)} duplicate basenames. New length {len(self._df)}"
            )
        if drop_graph_duplicates:
            old_len = len(self._df)
            self._df = self._df.drop_duplicates(subset=["info.decorated_graph_hash"], keep="first")
            logger.debug(
                f"Dropped {old_len - len(self._df)} duplicate graphs. New length {len(self._df)}"
            )

        # select by flavor
        self._df = self._df[self._df[f"flavor.{self._flavor}"]]

        self._df = self._df.reset_index(drop=True)

        if drop_nan:
            self._df.dropna(
                subset=[c for c in self._df.columns if c.startswith("features.")]
                + [c for c in self._df.columns if c.startswith("info.")],
                inplace=True,
            )
            self._df.reset_index(drop=True, inplace=True)

        if subset is not None:
            self._df = self._df.iloc[subset]
            self._df = self._df.reset_index(drop=True)

        self._structures = [
            os.path.join(self._structure_dir, f + ".cif") for f in self._df["info.qmof_id"]
        ]

        check_all_file_exists(self._structures)

        self._years = self._df["info.year"].values

        self._decorated_graph_hashes = self._df["info.decorated_graph_hash"].values
        self._undecorated_graph_hashes = self._df["info.undecorated_graph_hash"].values
        self._decorated_scaffold_hashes = self._df["info.decorated_scaffold_hash"].values
        self._undecorated_scaffold_hashes = self._df["info.undecorated_scaffold_hash"].values
        self._densities = self._df["info.density"].values
        self._labelnames = (c for c in self._df.columns if c.startswith("outputs."))
        self._featurenames = (c for c in self._df.columns if c.startswith("features."))
        self._infonames = (c for c in self._df.columns if c.startswith("info."))

    def get_subset(self, indices: Collection[int]) -> "AbstractStructureDataset":
        """Get a subset of the dataset.

        Args:
            indices (Collection[int]): indices of the structures to include.

        Returns:
            AbstractStructureDataset: a new dataset containing only the structures
                specified by the indices.
        """
        return QMOFDataset(
            version=self.version,
            drop_basename_duplicates=self._drop_basename_duplicates,
            drop_graph_duplicates=self._drop_graph_duplicates,
            subset=indices,
            flavor=self._flavor,
            drop_nan=self._drop_nan,
        )

    @property
    def available_info(self) -> Tuple[str]:
        return self._infonames

    @property
    def available_features(self) -> Tuple[str]:
        return self._featurenames

    @property
    def available_labels(self) -> Tuple[str]:
        return self._labelnames

    def get_labels(self, idx: Collection[int], labelnames: Collection[str] = None) -> np.ndarray:
        labelnames = labelnames if labelnames is not None else self._labelnames
        return self._df.iloc[idx][list(labelnames)].values

    @property
    def citations(self) -> Tuple[str]:
        return [
            "@article{Rosen2021,"
            "doi = {10.1016/j.matt.2021.02.015},"
            "url = {https://doi.org/10.1016/j.matt.2021.02.015},"
            "year = {2021},"
            "month = may,"
            "publisher = {Elsevier {BV}},"
            "volume = {4},"
            "number = {5},"
            "pages = {1578--1597},"
            "author = {Andrew S. Rosen and Shaelyn M. Iyer and Debmalya Ray "
            "and Zhenpeng Yao and Al{'{a}}n Aspuru-Guzik and Laura Gagliardi "
            "and Justin M. Notestein and Randall Q. Snurr},"
            "title = {Machine learning the quantum-chemical properties of"
            "metal{\textendash}organic frameworks for accelerated materials discovery},"
            "journal = {Matter}"
            "}",
            "@article{Rosen2022",
            "title={High-throughput predictions of metal--organic framework electronic properties:"
            " theoretical challenges, graph neural networks, and data exploration},"
            "author={Rosen, Andrew S and Fung, Victor and Huck, Patrick and O’Donnell, "
            "Cody T and Horton, Matthew K and Truhlar, Donald G and Persson, Kristin A "
            "and Notestein, Justin M and Snurr, Randall Q},"
            "journal={npj Computational Materials},"
            "volume={8},"
            "pages={112},"
            "year={2022},"
            "publisher={Nature Publishing Group}"
            "}",
        ]



# -*- coding: utf-8 -*-
"""Test the PartialChargeStats featurizer."""
from mofdscribe.featurizers.chemistry.partialchargestats import PartialChargeStats

from ..helpers import is_jsonable


def test_partial_charge_stats(hkust_structure, irmof_structure):
    """Make sure that the featurization works for typical MOFs and the number of
    features is as expected.
    """
    for structure in [hkust_structure, irmof_structure]:
        featurizer = PartialChargeStats()
        feats = featurizer.featurize(structure)
        assert len(feats) == 3
    assert len(featurizer.feature_labels()) == 3
    assert len(featurizer.citations()) == 2
    assert is_jsonable(dict(zip(featurizer.feature_labels(), feats)))
    assert feats.ndim == 1



# -*- coding: utf-8 -*-
"""Featurizers using persistent homology -- vectorized using Gaussian mixture models."""
from collections import defaultdict
from typing import List, Optional, Tuple, Union

import numpy as np
from loguru import logger
from pervect import PersistenceVectorizer
from pymatgen.core import IMolecule, IStructure, Molecule, Structure

from mofdscribe.featurizers.base import MOFBaseFeaturizer
from mofdscribe.featurizers.utils.extend import (
    operates_on_imolecule,
    operates_on_istructure,
    operates_on_molecule,
    operates_on_structure,
)

from ._tda_helpers import get_diagrams_for_structure


def _apply_and_fill(transformer_func, diagrams):
    applicable_diagrams = []
    rows_to_be_filled = []
    ok_rows = []
    for i, diagram in enumerate(diagrams):
        if diagram.shape[0] > 0:
            applicable_diagrams.append(diagram)
            ok_rows.append(i)
        else:
            rows_to_be_filled.append(i)
    results = transformer_func(applicable_diagrams)
    complete_results = np.zeros((len(diagrams), results.shape[1]))
    complete_results[ok_rows, :] = results
    if len(complete_results) != len(diagrams):
        raise ValueError(
            "Unexpected error. Number of feature vectors is not equal to number of structures"
        )
    return complete_results


def _fit_transform_structures(
    transformers,
    structures,
    atom_types: Tuple[str],
    compute_for_all_elements: bool,
    min_size: int,
    periodic: bool = False,
    no_supercell: bool = False,
    alpha_weight: Optional[str] = None,
):
    logger.info(f"Computing diagrams for {len(structures)} structures")
    diagrams = defaultdict(lambda: defaultdict(list))
    for structure in structures:
        res = get_diagrams_for_structure(
            structure,
            atom_types,
            compute_for_all_elements,
            min_size,
            periodic=periodic,
            no_supercell=no_supercell,
            alpha_weighting=alpha_weight,
        )
        for element, element_d in res.items():
            for dim, dim_d in element_d.items():
                diagrams[element][dim].append(dim_d)

    results = defaultdict(lambda: defaultdict(list))
    for element, element_transformers in transformers.items():
        for dim, transformer in element_transformers.items():
            if len(diagrams[element][dim]) == 0:
                raise ValueError(f"{element} dimension {dim} has no diagrams")
            try:

                results[element][dim] = _apply_and_fill(
                    transformer.fit_transform, diagrams[element][dim]
                )

                if not len(results[element][dim]) == len(structures):
                    raise ValueError(
                        "Unexpected error. Number of feature vectors is not equal to number of structures"
                    )

            except Exception as e:
                logger.error(
                    f"Error fitting transformer: {element} {dim} {diagrams[element][dim]}", e
                )

    return transformers, results


def _transform_structures(
    transformers,
    structures,
    atom_types: Tuple[str],
    compute_for_all_elements: bool,
    min_size: int,
    periodic: bool = False,
    no_supercell: bool = False,
    alpha_weight: Optional[str] = None,
):
    diagrams = defaultdict(lambda: defaultdict(list))
    for structure in structures:
        res = get_diagrams_for_structure(
            structure,
            atom_types,
            compute_for_all_elements,
            min_size,
            periodic=periodic,
            no_supercell=no_supercell,
            alpha_weighting=alpha_weight,
        )
        for element, element_d in res.items():
            for dim, dim_d in element_d.items():
                diagrams[element][dim].append(dim_d)

    results = defaultdict(lambda: defaultdict(list))

    for element, element_transformers in transformers.items():
        for dim, transformer in element_transformers.items():

            results[element][dim] = _apply_and_fill(
                transformer.fit_transform, diagrams[element][dim]
            )

    return results


@operates_on_imolecule
@operates_on_molecule
@operates_on_istructure
@operates_on_structure
class PHVect(MOFBaseFeaturizer):
    """Vectorizer for Persistence Diagrams (PDs) using Gaussian mixture models.

    The vectorization of a diagram is then the weighted maximum likelihood
    estimate of the mixture weights for the learned components given the diagram.

    Importantly, the vectorizations can still be used to compute approximate
    Wasserstein distances.
    """

    def __init__(
        self,
        atom_types: Tuple[str] = (
            "C-H-N-O",
            "F-Cl-Br-I",
            "Cu-Mn-Ni-Mo-Fe-Pt-Zn-Ca-Er-Au-Cd-Co-Gd-Na-Sm-Eu-Tb-V"
            "-Ag-Nd-U-Ba-Ce-K-Ga-Cr-Al-Li-Sc-Ru-In-Mg-Zr-Dy-W-Yb-Y-"
            "Ho-Re-Be-Rb-La-Sn-Cs-Pb-Pr-Bi-Tm-Sr-Ti-Hf-Ir-Nb-Pd-Hg-"
            "Th-Np-Lu-Rh-Pu",
        ),
        compute_for_all_elements: bool = True,
        dimensions: Tuple[int] = (1, 2),
        min_size: int = 20,
        n_components: int = 20,
        apply_umap: bool = False,
        umap_n_components: int = 2,
        umap_metric: str = "hellinger",
        p: int = 1,
        random_state: Optional[int] = None,
        periodic: bool = False,
        no_supercell: bool = False,
        primitive: bool = False,
        alpha_weight: Optional[str] = None,
    ) -> None:
        """Construct a PHVect instance.

        Args:
            atom_types (tuple): Atoms that are used to create substructures
                that are analysed using persistent homology.
                If multiple atom types separated by hash are provided,
                e.g. "C-H-N-O", then the substructure consists of all atoms of type C, H, N, or O.
                Defaults to ( "C-H-N-O", "F-Cl-Br-I",
                "Cu-Mn-Ni-Mo-Fe-Pt-Zn-Ca-Er-Au-Cd-Co-Gd-Na-Sm-Eu-Tb-V-Ag-Nd-U-Ba-Ce-K-Ga-
                Cr-Al-Li-Sc-Ru-In-Mg-Zr-Dy-W-Yb-Y-Ho-Re-Be-Rb-La-Sn-Cs-Pb-Pr-Bi-Tm-Sr-Ti-
                Hf-Ir-Nb-Pd-Hg-Th-Np-Lu-Rh-Pu", ).
            compute_for_all_elements (bool): If true, compute persistence images
                for full structure (i.e. with all elements). If false, it will only do it
                for the substructures specified with `atom_types`. Defaults to True.
            dimensions (Tuple[int]): Dimensions of topological features to consider
                for persistence images. Defaults to (1, 2).
            min_size (int): Minimum supercell size (in Angstrom).
                Defaults to 20.
            n_components (int): The number of components or dimensions to use
                in the vectorized representation. Defaults to 20.
            apply_umap (bool):  Whether to apply UMAP to the results to generate
                a low dimensional Euclidean space representation of the diagrams.
                Defaults to False.
            umap_n_components (int):  The number of dimensions of euclidean space
                to use when representing the diagrams via UMAP. Defaults to 2.
            umap_metric (str):  What metric to use for the UMAP embedding if ``apply_umap`` is enabled
                (this option will be ignored if ``apply_umap`` is ``False``).
                Should be one of:
                * ``"wasserstein"``
                * ``"hellinger"``
                Note that if ``"wasserstein"`` is used then transforming new data will not be possible.
                Defaults to "hellinger".
            p (int):  The p in the p-Wasserstein distance to compute.
            random_state (_type_): random state propagated to the Gaussian mixture models (and UMAP).
                Defaults to None.
            periodic (bool): If true, then periodic Euclidean is used in the analysis (experimental!).
                Defaults to False.
            no_supercell (bool): If true, then the supercell is not created.
                Defaults to False.
            primitive (bool): If True, the structure is reduced to its primitive
                form before the descriptor is computed. Defaults to False.
            alpha_weight (Optional[str]): If specified, the use weighted alpha shapes,
                i.e., replacing the points with balls of varying radii.
                For instance `atomic_radius_calculated` or `van_der_waals_radius`.
        """
        atom_types = [] if atom_types is None else atom_types
        self.elements = atom_types
        self.atom_types = (
            list(atom_types) + ["all"] if compute_for_all_elements else list(atom_types)
        )
        self.compute_for_all_elements = compute_for_all_elements
        self.min_size = min_size
        self.dimensions = dimensions
        self.transformers = defaultdict(lambda: defaultdict(dict))
        for atom_type in self.atom_types:
            for dim in self.dimensions:
                self.transformers[atom_type][f"dim{dim}"] = PersistenceVectorizer(
                    n_components=n_components,
                    apply_umap=apply_umap,
                    umap_n_components=umap_n_components,
                    umap_metric=umap_metric,
                    p=p,
                    random_state=random_state,
                )
        self.n_components = n_components
        self.apply_umap = apply_umap
        self.umap_n_components = umap_n_components
        self.umap_metric = umap_metric
        self.random_state = random_state
        self._fitted = False
        self.periodic = periodic
        self.no_supercell = no_supercell
        self.alpha_weight = alpha_weight
        super().__init__(primitive=primitive)

    def _get_feature_labels(self) -> List[str]:
        labels = []
        for atom_type in self.atom_types:
            for dim in self.dimensions:
                for i in range(self.n_components):
                    labels.append(f"phvect_{atom_type}_{dim}_{i}")

        return labels

    def feature_labels(self) -> List[str]:
        return self._get_feature_labels()

    def _featurize(
        self, structure: Union[Structure, IStructure, Molecule, IMolecule]
    ) -> np.ndarray:
        if not self._fitted:
            raise ValueError("Must call fit before featurizing")
        res = _transform_structures(
            self.transformers,
            [structure],
            self.elements,
            self.compute_for_all_elements,
            self.min_size,
            periodic=self.periodic,
            no_supercell=self.no_supercell,
            alpha_weight=self.alpha_weight,
        )
        compiled_results = self._reshape_results(res, 1).flatten()
        return compiled_results

    def fit(self, structures: Union[Structure, IStructure, Molecule, IMolecule]) -> "PHVect":
        if self.primitive:
            structures = self._get_primitive_many(structures)
        self.transformers, _ = _fit_transform_structures(
            self.transformers,
            structures,
            self.elements,
            self.compute_for_all_elements,
            self.min_size,
            periodic=self.periodic,
            no_supercell=self.no_supercell,
            alpha_weight=self.alpha_weight,
        )
        self._fitted = True
        return self

    def _reshape_results(self, results, num_structures) -> np.ndarray:
        compiled_results = np.zeros((num_structures, len(self._get_feature_labels())))
        n_col = 0
        for _, element_results in results.items():
            for _, result in element_results.items():
                compiled_results[:, n_col : n_col + self.n_components] = result
                n_col += self.n_components
        return compiled_results

    def fit_transform(
        self, structures: Union[Structure, IStructure, Molecule, IMolecule]
    ) -> np.ndarray:
        if self.primitive:
            structures = self._get_primitive_many(structures)
        self.transformers, results = _fit_transform_structures(
            self.transformers,
            structures,
            self.elements,
            self.compute_for_all_elements,
            self.min_size,
            periodic=self.periodic,
            no_supercell=self.no_supercell,
            alpha_weight=self.alpha_weight,
        )
        compiled_results = self._reshape_results(results, len(structures))
        self._fitted = True
        return compiled_results

    def citations(self):
        return [
            "@article{perea2019approximating,"
            "title   = {Approximating Continuous Functions on Persistence "
            "Diagrams Using Template Functions},"
            "author  = {Jose A. Perea and Elizabeth Munch and Firas A. Khasawneh},"
            "year    = {2019},"
            "journal = {arXiv preprint arXiv: Arxiv-1902.07190}"
            "}",
            "@article{tymochko2019adaptive,"
            "title   = {Adaptive Partitioning for Template Functions on Persistence Diagrams},"
            "author  = {Sarah Tymochko and Elizabeth Munch and Firas A. Khasawneh},"
            "year    = {2019},"
            "journal = {arXiv preprint arXiv: Arxiv-1910.08506}"
            "}",
            "@article{doi:10.1021/acs.jpcc.0c01167,"
            "author = {Krishnapriyan, Aditi S. and Haranczyk, Maciej and Morozov, Dmitriy},"
            "title = {Topological Descriptors Help Predict Guest Adsorption in Nanoporous Materials},"
            "journal = {The Journal of Physical Chemistry C},"
            "volume = {124},"
            "number = {17},"
            "pages = {9360-9368},"
            "year = {2020},"
            "doi = {10.1021/acs.jpcc.0c01167},"
            "}",
        ]

    def implementors(self):
        return ["Kevin Maik Jablonka", "Aditi Krishnapriyan"]



# -*- coding: utf-8 -*-
"""Helpers for computing histograms."""
import math
from typing import Optional

import numpy as np
from scipy.stats import gaussian_kde


def get_rdf(
    array: np.array,
    lower_lim: float,
    upper_lim: float,
    bin_size: float,
    num_sites: Optional[int] = None,
    volume: Optional[float] = None,
    normalized: bool = True,
    density: bool = False,
) -> np.array:
    """Compute the RDF

    Args:
        array (np.array): distances
        lower_lim (float): lower limit of the RDF
        upper_lim (float): upper limit of the RDF
        bin_size (float): size of the bins
        num_sites (int, optional): number of sites in the cell, used for normalization
        volume (float, optional): volume of the cell, used for normalization
        normalized (bool): If True, normalize the RDF. Defaults to True.
        density (bool): If True, return the density of the RDF. Defaults to False.

    Returns:
        np.array: RDF
    """
    dist_hist, dist_bins = np.histogram(
        array,
        bins=np.arange(lower_lim, upper_lim + bin_size, bin_size),
        density=density,
    )

    if normalized:
        shell_vol = 4.0 / 3.0 * math.pi * (np.power(dist_bins[1:], 3) - np.power(dist_bins[:-1], 3))
        number_density = num_sites / volume
        rdf = dist_hist / shell_vol / number_density
        return rdf

    return dist_hist.astype(np.float64)


def smear_histogram(histogram: np.array, bw: float, lower_lim: float, upper_lim: float) -> np.array:
    """Use a gaussian kernel to smooth the histogram.

    Args:
        histogram (np.array): histogram to be smoothed
        bw (float): bandwidth of the gaussian kernel
        lower_lim (float): lower limit of the RDF
        upper_lim (float): upper limit of the RDF

    Returns:
        np.array: smoothed histogram
    """
    kernel = gaussian_kde(histogram, bw_method=bw)
    x = np.linspace(lower_lim, upper_lim, len(histogram))
    y = kernel(x)
    return y



# -*- coding: utf-8 -*-
"""Sanity checks for datasets."""
import os
from typing import List, Union

import pandas as pd

__all__ = ["length_check", "check_all_file_exists"]


def length_check(df: pd.DataFrame, expected_length: int) -> None:
    """Ensure the length of the dataframe is as expected."""
    if not len(df) == expected_length:
        raise ValueError(
            "Length of dataframe does not match expected length. Found {}, expected {}".format(
                len(df), expected_length
            )
        )


def check_all_file_exists(filelist: List[Union[str, os.PathLike]]) -> None:
    """Ensure that all expected structures have been downloaded."""
    for f in filelist:
        if not os.path.exists(f):
            raise ValueError(f"File {f} does not exist.")



# -*- coding: utf-8 -*-
"""Test splitters."""
from collections import Counter

import numpy as np
import pytest

from mofdscribe.datasets.core_dataset import CoREDataset
from mofdscribe.splitters.splitters import (
    LOCOCV,
    BaseSplitter,
    ClusterSplitter,
    ClusterStratifiedSplitter,
    DensitySplitter,
    HashSplitter,
    KennardStoneSplitter,
    TimeSplitter,
)

FEATURES = list(CoREDataset().available_features)


@pytest.mark.xdist_group(name="core-ds")
@pytest.mark.parametrize("sample_frac", (0.2, 0.5, 1.0))
@pytest.mark.parametrize("shuffle", (True, False))
def test_base_splitter(sample_frac, shuffle):
    ds = CoREDataset()
    group_pool = ["A", "B", "C"]
    groups = np.random.choice(group_pool, size=len(ds))

    class MySplitter(BaseSplitter):
        def _get_groups(self):
            return groups

    splitter = MySplitter(ds, sample_frac=sample_frac, shuffle=shuffle)

    train_idx, test_idx = splitter.train_test_split()

    if sample_frac < 1:
        assert len(train_idx) + len(test_idx) < len(ds)
    groups_train = set(groups[train_idx])
    groups_test = set(groups[test_idx])
    assert groups_train & groups_test == set()


@pytest.mark.xdist_group(name="core-ds")
@pytest.mark.parametrize("sample_frac", (0.2, 0.5, 1.0))
def test_hash_splitter(sample_frac):
    """Ensure that the splits add up to the total number of structures and are non-overlapping."""
    ds = CoREDataset()
    hs = HashSplitter(ds, hash_type="undecorated_scaffold_hash", sample_frac=sample_frac)

    for train, test in hs.k_fold(k=5):

        assert set(train) & set(test) == set()

        if sample_frac < 1:
            assert len(train) + len(test) < len(ds)
        else:
            assert len(set(list(train) + list(test))) == len(ds)

    splits = hs.train_valid_test_split(frac_train=0.5, frac_valid=0.3)
    assert len(splits) == 3
    assert len(splits[0]) > len(splits[1]) > len(splits[2])


@pytest.mark.xdist_group(name="core-ds")
def test_time_splitter():
    """Ensure that the splits add up to the total number of structures and are non-overlapping."""
    ds = CoREDataset()
    ts = TimeSplitter(ds)

    for train, test in ts.k_fold(k=5):

        assert set(train) & set(test) == set()
        assert len(set(list(train) + list(test))) == len(ds)

    splits = ts.train_valid_test_split(frac_train=0.5, frac_valid=0.3)
    assert len(splits) == 3
    assert len(splits[0]) > len(splits[1]) > len(splits[2])


@pytest.mark.xdist_group(name="core-ds")
def test_density_splitter():
    """Ensure that the splits add up to the total number of structures and are non-overlapping."""
    ds = CoREDataset()

    dens_splitter = DensitySplitter(ds)
    for train, test in dens_splitter.k_fold(k=5):
        assert all(dens_splitter._grouping_q == np.linspace(0, 1, 6))
        assert set(train) & set(test) == set()
        assert len(set(list(train) + list(test))) == len(ds)

    splits = dens_splitter.train_valid_test_split(frac_train=0.5, frac_valid=0.3)
    assert isinstance(splits, tuple)
    assert all(dens_splitter._grouping_q == np.linspace(0, 1, 4))
    assert len(splits) == 3
    assert len(splits[0]) > len(splits[1]) >= len(splits[2])

    groups = dens_splitter._get_groups()

    set0 = set(groups[splits[0]])
    set1 = set(groups[splits[1]])
    set2 = set(groups[splits[2]])

    assert set0 & set1 == set()
    assert set1 & set2 == set()
    assert set0 & set2 == set()

    splits = dens_splitter.train_test_split(
        frac_train=0.5,
    )
    assert isinstance(splits, tuple)
    assert all(dens_splitter._grouping_q == np.linspace(0, 1, 3))
    groups = dens_splitter._get_groups()

    set0 = set(groups[splits[0]])
    set1 = set(groups[splits[1]])


@pytest.mark.xdist_group(name="core-ds")
def test_kennard_stone_splitter():
    """Ensure that the splits add up to the total number of structures and are non-overlapping."""
    ds = CoREDataset()
    # To make sure that the test does not take too long, we only use a small subset of the dataset.
    ds._df = ds._df.iloc[:500]
    ds._structures = ds._structures[:500]

    fps = KennardStoneSplitter(
        ds,
        feature_names=FEATURES,
    )
    for train, test in fps.k_fold(k=5):
        assert set(train) & set(test) == set()
        assert len(set(list(train) + list(test))) == len(ds)

    splits = fps.train_valid_test_split(frac_train=0.5, frac_valid=0.3)
    assert len(splits) == 3
    assert len(splits[0]) > len(splits[1]) > len(splits[2])


@pytest.mark.xdist_group(name="core-ds")
def test_cluster_splitter():
    """Ensure that the splits add up to the total number of structures and are non-overlapping."""
    ds = CoREDataset()
    # To make sure that the test does not take too long, we only use a small subset of the dataset.
    ds._df = ds._df.iloc[:500]
    ds._structures = ds._structures[:500]

    fps = ClusterSplitter(ds, feature_names=FEATURES, n_pca_components=50)
    for train, test in fps.k_fold(k=5):

        assert len(set(list(train) + list(test))) == len(ds)
        assert set(train) & set(test) == set()

    splits = fps.train_valid_test_split(frac_train=0.5, frac_valid=0.3)
    assert len(splits) == 3
    assert len(splits[0]) > len(splits[1]) > len(splits[2])


@pytest.mark.xdist_group(name="core-ds")
def test_cluster_stratified_splitter():
    """Ensure that the splits add up to the total number of structures and are non-overlapping.

    Also make sure that the class ratios are indeed approximately stratified.
    """
    ds = CoREDataset()

    fps = ClusterStratifiedSplitter(ds, feature_names=FEATURES, n_clusters=2, n_pca_components=50)
    for train, test in fps.k_fold(k=5):

        assert len(set(list(train) + list(test))) == len(ds)
        assert set(train) & set(test) == set()

        # also make sure that the class ratios roughly make sense
        train_groups = fps._stratification_groups[train]
        test_groups = fps._stratification_groups[test]

        assert len(set(train_groups)) == len(set(test_groups)) == 2
        train_counter = Counter(train_groups)
        test_counter = Counter(test_groups)
        assert np.abs(train_counter[0] / train_counter[1] - test_counter[0] / test_counter[1]) < 10

    splits = fps.train_valid_test_split(frac_train=0.5, frac_valid=0.3)
    assert len(splits) == 3
    assert len(splits[0]) > len(splits[1]) > len(splits[2])


@pytest.mark.xdist_group(name="core-ds")
def test_locov():
    """Ensure that the splits add up to the total number of structures and are non-overlapping.

    Also make sure that the class ratios are indeed approximately stratified.
    """
    ds = CoREDataset()
    # To make sure that the test does not take too long, we only use a small subset of the dataset.
    ds._df = ds._df.iloc[:1000]
    ds._structures = ds._structures[:1000]

    fps = LOCOCV(ds, feature_names=FEATURES, n_pca_components=100)
    for train, test in fps.k_fold(k=5):

        assert len(set(list(train) + list(test))) == len(ds)
        assert set(train) & set(test) == set()
        assert len(train) > len(test)
    splits = fps.train_valid_test_split()
    assert len(splits) == 3
    assert len(splits[0]) > len(splits[2]) > len(splits[1])



# -*- coding: utf-8 -*-
"""Helper functions for computing metrics."""
from .adverserial import AdverserialValidator  # noqa: F401
from .regression import RegressionMetrics, get_regression_metrics, top_n_in_top_k  # noqa: F401



# -*- coding: utf-8 -*-
from .splitters import ClusterSplitter  # noqa: F401
from .splitters import ClusterStratifiedSplitter  # noqa: F401
from .splitters import DensitySplitter  # noqa: F401
from .splitters import HashSplitter  # noqa: F401
from .splitters import KennardStoneSplitter  # noqa: F401
from .splitters import TimeSplitter  # noqa: F401



# -*- coding: utf-8 -*-
"""Compute features on the BUs and then aggregate them."""
from typing import Collection, List, Optional, Tuple, Union

import numpy as np
from matminer.featurizers.base import BaseFeaturizer
from pydantic import BaseModel
from pymatgen.core import IMolecule, IStructure, Molecule, Structure

from mofdscribe.featurizers.bu.utils import boxed_molecule
from mofdscribe.featurizers.utils import nan_array, set_operates_on
from mofdscribe.featurizers.utils.aggregators import ARRAY_AGGREGATORS


class MOFBBs(BaseModel):
    """Container for MOF building blocks."""

    nodes: Optional[List[Union[Structure, Molecule, IStructure, IMolecule]]]
    linkers: Optional[List[Union[Structure, Molecule, IStructure, IMolecule]]]


# ToDo: Support `MultipleFeaturizer`s (should be ok, if we recursively call the operates_on method).
class BUFeaturizer(BaseFeaturizer):
    """
    Compute features on the BUs and then aggregate them.

    This can be useful if you want to compute some features on the BUs
    and them aggregrate them to obtain one fixed-length feature vector for the MOF.

    .. warning::
        Note that, currently, not all featurizers can operate on both
        Structures and Molecules.
        If you want to include featurizers that can only operate on one type
        (e.g. :py:obj:`~mofdscribe.bu.rdkitadaptor.RDKitAdaptor`
        and :py:obj:`~mofdscribe.chemistry.amd.AMD`)
        then you need to create two separate MOFBBs and BUFeaturizer objects.

    Examples:
        >>> from mofdscribe.bu import BUFeaturizer, MOFBBs
        >>> from mofdscribe.bu.rdkitadaptor import RDKitAdaptor
        >>> from rdkit.Chem.Descriptors3D import Asphericity
        >>> from pymatgen.core import Molecule
        >> from pymatgen.io.babel import BabelMolAdaptor
        >>> base_featurizer = RDKitAdaptor(featurizer=Asphericity, feature_labels=["asphericity"])
        >>> bu_featurizer = BUFeaturizer(featurizer=base_featurizer, aggregations=("mean", "std", "min", "max"))
        >>> bu_featurizer.featurize(
                mofbbs=MOFBBs(nodes=[BabelMolAdaptor.from_string(
                    "[CH-]1C=CC=C1.[CH-]1C=CC=C1.[Fe+2]", "smi").pymatgen_mol],
                linkers=[BabelMolAdaptor.from_string("CCCC", "smi").pymatgen_mol]))

    """

    def __init__(
        self, featurizer: BaseFeaturizer, aggregations: Tuple[str] = ("mean", "std", "min", "max")
    ) -> None:
        """
        Construct a new BUFeaturizer.

        Args:
            featurizer (BaseFeaturizer): The featurizer to use.
                Currently, we do not support `MultipleFeaturizer`s.
                Please, instead, use multiple BUFeaturizers.
                If you use a featurizer that is not implemented in mofdscribe
                (e.g. a matminer featurizer), you need to wrap using a method
                that describes on which data objects the featurizer can operate on.
                If you do not do this, we default to assuming that it operates on structures.
            aggregations (Tuple[str]): The aggregations to use.
                Must be one of :py:obj:`ARRAY_AGGREGATORS`.
        """
        self._featurizer = featurizer
        self._aggregations = aggregations
        set_operates_on(self, featurizer)

    def feature_labels(self) -> List[str]:
        labels = []
        base_labels = self._featurizer.feature_labels()
        for bb in ["node", "linker"]:
            for aggregation in self._aggregations:
                for label in base_labels:
                    labels.append(f"{bb}_{aggregation}_{label}")
        return labels

    def _extract_bbs(
        self,
        structure: Optional[Union[Structure, IStructure]] = None,
        mofbbs: Optional[MOFBBs] = None,
    ):
        if structure is None and mofbbs is None:
            raise ValueError("You must provide a structure or mofbbs.")

        if structure is not None:
            from moffragmentor import MOF

            mof = MOF.from_structure(structure)
            fragments = mof.fragment()

            if self._operates_on in ("both", "molecule"):
                linkers = [linker.molecule for linker in fragments.linkers]
                nodes = [node.molecule for node in fragments.nodes]
            else:
                # create a boxed structure
                linkers = [boxed_molecule(linker.molecule) for linker in fragments.linkers]
                nodes = [boxed_molecule(node.molecule) for node in fragments.nodes]

        if mofbbs is not None:

            linkers = list(mofbbs.linkers) if mofbbs.linkers is not None else []
            nodes = list(mofbbs.nodes) if mofbbs.nodes is not None else []
            types = [type(node) for node in nodes] + [type(linker) for linker in linkers]

            if not len(set(types)) == 1:
                raise ValueError("All nodes and linkers must be of the same type.")

            this_type = types[0]
            if this_type in (Structure, IStructure) and self._operates_on in ("both", "structure"):
                # this is the simple case, we do not need to convert to molecules
                pass
            elif this_type in (Molecule, IMolecule) and self._operates_on in ("both", "molecule"):
                # again simple case, we do not need to convert to structures

                pass
            elif this_type in (Molecule, IMolecule) and (self._operates_on == "structure"):
                # we need to convert to structures
                nodes = [boxed_molecule(node) for node in nodes]
                linkers = [boxed_molecule(linker) for linker in linkers]
            elif this_type in (Structure, IStructure) and (self._operates_on == "molecule"):
                raise ValueError(
                    "You provided structures for a featurizer that operates on molecules. "
                    / "Cannot automatically convert to molecules from structures."
                )
            else:
                raise RuntimeError("Unexpected type of nodes or linkers.")

        return nodes, linkers

    def fit(
        self,
        structures: Optional[Collection[Structure]] = None,
        mofbbs: Optional[Collection[MOFBBs]] = None,
    ) -> None:
        """
        Fit the featurizer to the given structures.

        Args:
            structures (Collection[Structure], optional): The structures to featurize.
            mofbbs (Collection[MOFBBs], optional): The MOF fragments (nodes and linkers).
        """
        all_nodes, all_linkers = [], []
        if structures is not None:
            for structure in structures:
                nodes, linkers = self._extract_bbs(structure=structure)
                all_nodes.extend(nodes)
                all_linkers.extend(linkers)
        if mofbbs is not None:
            for mofbb in mofbbs:
                nodes, linkers = self._extract_bbs(mofbbs=mofbb)
                all_nodes.extend(nodes)
                all_linkers.extend(linkers)

        all_fragments = all_nodes + all_linkers
        self._featurizer.fit(all_fragments)

    # ToDo:
    # - Perhaps use type dispatch instead of different keyword arguments.
    #   (we can use fastcore or code it ourselves)
    # - We can also directly pass the graph to the featurizer if it want to work
    #     on the graph.
    def featurize(
        self,
        structure: Optional[Union[Structure, IStructure]] = None,
        mofbbs: Optional[MOFBBs] = None,
    ) -> np.ndarray:
        """
        Compute features on the BUs and then aggregate them.

        If you provide a structure, we will fragment the MOF into BUs.
        If you already have precomputed fragements or only want to consider a subset
        of the BUs, you can provide them manually via the `mofbbs` argument.

        If you manually provide the `mofbbs`,  we will convert molecules to structures
        where possible.

        Args:
            structure (Union[Structure, IStructure], optional): The structure to featurize.
            mofbbs (MOFBBs, optional): The MOF fragments (nodes and linkers).

        Returns:
            A numpy array of features.
        """
        # if i know what the featurizer wants, I can always cast to a structure
        num_features = len(self._featurizer.feature_labels())
        nodes, linkers = self._extract_bbs(structure, mofbbs)
        linker_feats = [self._featurizer.featurize(linker) for linker in linkers]
        if not linker_feats:
            linker_feats = [nan_array(num_features)]

        node_feats = [self._featurizer.featurize(node) for node in nodes]
        if not node_feats:
            node_feats = [nan_array(num_features)]

        aggregated_linker_feats = []
        for aggregation in self._aggregations:
            aggregated_linker_feats.extend(ARRAY_AGGREGATORS[aggregation](linker_feats, axis=0))
        aggregated_linker_feats = np.array(aggregated_linker_feats)

        aggregated_node_feats = []
        for aggregation in self._aggregations:
            aggregated_node_feats.extend(ARRAY_AGGREGATORS[aggregation](node_feats, axis=0))
        aggregated_node_feats = np.array(aggregated_node_feats)

        return np.concatenate((aggregated_node_feats, aggregated_linker_feats))

    def citations(self) -> List[str]:
        return self._featurizer.citations()

    def implementors(self) -> List[str]:
        return self._featurizer.implementors()



# -*- coding: utf-8 -*-
"""Utils for MOFDescribe."""

import sys
from shutil import which

import numpy as np
from pymatgen.core import Molecule, Structure

if sys.version_info.major == 3 and sys.version_info.minor >= 10:
    from collections.abc import MutableMapping
else:
    from collections import MutableMapping


def nan_array(size):
    return np.full(size, np.nan)


def is_tool(name: str) -> bool:
    """Check whether `name` is on PATH and marked as executable.

    Taken from https://stackoverflow.com/questions/11210104/check-if-a-program-exists-from-a-python-script

    Args:
        name (str): executable name

    Returns:
        bool: True if the executable in PATH
    """
    return which(name) is not None


def flatten(d: dict, parent_key: str = "", sep: str = ".") -> dict:
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, MutableMapping):
            items.extend(flatten(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


def set_operates_on(self, featurizer):
    try:
        _operates_on = featurizer.operates_on()
        if (Structure in _operates_on) and (Molecule in _operates_on):
            self._operates_on = "both"
        elif Structure in _operates_on:
            self._operates_on = "structure"
        elif Molecule in _operates_on:
            self._operates_on = "molecule"

    except AttributeError:
        self._operates_on = "structure"



# -*- coding: utf-8 -*-
"""Test fixtures"""
import json
import os
from glob import glob

import pytest
from pymatgen.analysis.graphs import MoleculeGraph
from pymatgen.core import IStructure, Molecule, Structure
from structuregraph_helpers.create import get_structure_graph

THIS_DIR = os.path.dirname(os.path.abspath(__file__))


@pytest.fixture(scope="session")
def hkust_structure():
    """Return a pymatgen Structure for HKUST"""
    return IStructure.from_file(os.path.join(THIS_DIR, "test_files", "HKUST-1.cif"))


@pytest.fixture(scope="session")
def hkust_la_structure():
    """Return a pymatgen Structure for HKUST"""
    return IStructure.from_file(os.path.join(THIS_DIR, "test_files", "HKUST-1-La.cif"))


@pytest.fixture(scope="session")
def abacuf_structure():
    """Return a pymatgen Structure for ABACUF"""
    return IStructure.from_file(os.path.join(THIS_DIR, "test_files", "ABACUF.cif"))


@pytest.fixture(scope="session")
def irmof_structure():
    """Return a pymatgen Structure for IRMOF"""
    return IStructure.from_file(os.path.join(THIS_DIR, "test_files", "IRMOF-1.cif"))


@pytest.fixture(scope="session")
def cof_structure():
    """Return a pymatgen Structure for a COF"""
    return IStructure.from_file(os.path.join(THIS_DIR, "test_files", "20450N2_ddec.cif"))


@pytest.fixture(scope="session")
def hkust_structure_graph():
    """Return a pymatgen StructureGraph for HKUST"""
    return get_structure_graph(
        IStructure.from_file(os.path.join(THIS_DIR, "test_files", "HKUST-1.cif"))
    )


@pytest.fixture(scope="session")
def molecule_graph():
    """Return a pymatgen MoleculeGraph for a HKUST node"""
    with open(os.path.join(THIS_DIR, "test_files", "test_molecule_graph.json"), "r") as handle:
        mol_graph = MoleculeGraph.from_dict(json.loads(handle.read()))
    return mol_graph


@pytest.fixture(scope="session")
def molecule():
    """Return a pymatgen Molecule for a HKUST node"""
    with open(os.path.join(THIS_DIR, "test_files", "test_molecule.json"), "r") as handle:
        mol = Molecule.from_dict(json.loads(handle.read()))
    return mol


@pytest.fixture(scope="session")
def linker_molecule():
    """Return a pymatgen Molecule for BTC linker"""
    with open(os.path.join(THIS_DIR, "test_files", "linker_molecule.json"), "r") as handle:
        mol = Molecule.from_dict(json.loads(handle.read()))
    return mol


@pytest.fixture(scope="session")
def triangle_structure():
    """Return a pymatgen Structure for the `connecting sites` of a BTC linker"""
    with open(os.path.join(THIS_DIR, "test_files", "triangle_structure.json"), "r") as handle:
        s = IStructure.from_dict(json.loads(handle.read()))
    return s


@pytest.fixture(scope="session")
def floating_structure():
    """Return IRMOF with floating mol"""
    return IStructure.from_file(os.path.join(THIS_DIR, "test_files", "floating_check.cif"))


@pytest.fixture(scope="session")
def hkust_linker_structure():
    return Structure.from_file(os.path.join(THIS_DIR, "test_files", "BTC.cif"))


@pytest.fixture(scope="session")
def dataset_files():
    structures = glob(os.path.join(THIS_DIR, "test_files", "test_dataset", "structures", "*.cif"))
    dataset = glob(os.path.join(THIS_DIR, "test_files", "test_dataset", "data.json"))
    return structures, dataset


@pytest.fixture(scope="session")
def dataset_folder():
    return os.path.join(THIS_DIR, "test_files", "test_dataset")



# -*- coding: utf-8 -*-
"""Test the BU featurizer."""

import numpy as np
from matminer.featurizers.site import SOAP
from matminer.featurizers.structure import SiteStatsFingerprint
from pymatgen.core import Structure

from mofdscribe.featurizers.bu.bu_featurizer import BUFeaturizer, MOFBBs
from mofdscribe.featurizers.topology import PHStats


def test_bu_featurizer(hkust_structure, molecule):
    """Test that we can call BU featurizers with pymatgen molecules."""
    featurizer = BUFeaturizer(PHStats(no_supercell=True))

    mofbbs = MOFBBs(nodes=None, linkers=[molecule])
    features = featurizer.featurize(mofbbs=mofbbs)
    assert features.shape == (768,)
    assert features[600] >= 0
    assert features[600] < 2

    featurizer = BUFeaturizer(PHStats(no_supercell=True))
    features = featurizer.featurize(structure=hkust_structure)
    assert features.shape == (768,)
    assert features[0] > 0
    assert features[0] < 2


def test_bu_featurizer_with_matminer_featurizer(hkust_structure, hkust_linker_structure):
    """Test that we can call BU featurizers with matminer molecules."""
    # we disable the periodic keyword to be able to compare with the molecules
    base_feat = SiteStatsFingerprint(SOAP(6, 8, 8, 0.4, False, "gto", False))
    hkust_structure = Structure.from_sites(hkust_structure.sites)
    base_feat.fit([hkust_structure])
    featurizer = BUFeaturizer(base_feat, aggregations=("mean",))
    features = featurizer.featurize(structure=hkust_structure)
    assert features.shape == (2592 * 2,)
    assert features[0] >= 0
    assert features[0] < 2

    linker_feats = [f for f in featurizer.feature_labels() if "linker" in f]
    assert len(linker_feats) == 2592

    # test that our fit method works
    featurizer = BUFeaturizer(
        SiteStatsFingerprint(SOAP(6, 8, 8, 0.4, False, "gto", False)), aggregations=("mean",)
    )
    featurizer.fit([hkust_structure])
    features_direct_fit = featurizer.featurize(structure=hkust_structure)
    assert np.allclose(features, features_direct_fit, rtol=0.01)

    # test that the linker features are actually the ones we get when
    # we featurize the linker
    linker_feats = featurizer._featurizer.featurize(hkust_linker_structure)
    linker_feature_mask = [i for i, f in enumerate(featurizer.feature_labels()) if "linker" in f]
    assert np.allclose(features[linker_feature_mask], linker_feats, rtol=0.01, equal_nan=True)



# -*- coding: utf-8 -*-
"""Test the splitter helping functions."""

from random import shuffle

import numpy as np
import pytest

from mofdscribe.splitters.utils import (
    check_fraction,
    downsample_splits,
    get_train_valid_test_sizes,
    grouped_stratified_train_test_partition,
    grouped_train_valid_test_partition,
    kennard_stone_sampling,
    stratified_train_test_partition,
)


def test_kennard_stone_sampling():
    """Ensure we get the order we would expect."""
    X = np.array([[1, 2, 3], [4, 5, 6], [8, 8, 9]])  # noqa: N806

    indices = kennard_stone_sampling(X)
    assert indices == [2, 0, 1]

    # Make sure also the other options do not complain
    indices = kennard_stone_sampling(X, centrality_measure="median")
    assert indices == [2, 0, 1]

    indices = kennard_stone_sampling(X, centrality_measure="random")
    assert len(indices) == 3  # we cannot guarantee the order of the indices


@pytest.mark.parametrize("number_of_groups", [30, 50, 80, 500, 8_000])
@pytest.mark.parametrize("shuffle", [True])
def test_grouped_stratified_train_test_partition(number_of_groups, shuffle):
    # perhaps use hypothesis for fuzzing the data
    datasize = 10_000

    y = np.random.normal(0, 1, size=datasize)
    groups = np.random.choice(np.arange(number_of_groups), size=datasize)

    train_indices, valid_indices, test_indices = grouped_stratified_train_test_partition(
        y, groups, train_size=0.5, valid_size=0.25, test_size=0.25, shuffle=shuffle
    )

    assert len(train_indices) + len(test_indices) + len(valid_indices) == datasize

    # make sure we have no index overlap
    assert len(np.intersect1d(train_indices, valid_indices)) == 0
    assert len(np.intersect1d(train_indices, test_indices)) == 0

    # make sure we have no group overlap
    test_groups = groups[test_indices]
    train_groups = groups[train_indices]
    valid_groups = groups[valid_indices]
    assert set(test_groups).intersection(set(train_groups)) == set()
    assert set(test_groups).intersection(set(valid_groups)) == set()
    assert set(train_groups).intersection(set(valid_groups)) == set()

    # since we stratify we should have comparable medians
    train_mean = np.median(y[train_indices])
    valid_mean = np.median(y[valid_indices])
    test_mean = np.median(y[test_indices])
    assert (
        pytest.approx(train_mean, abs=0.1)
        == pytest.approx(valid_mean, abs=0.1)
        == pytest.approx(test_mean, abs=0.1)
    )

    # now, no validation set
    train_indices, _, test_indices = grouped_stratified_train_test_partition(
        y, groups, train_size=0.8, valid_size=0.0, test_size=0.2, shuffle=shuffle
    )

    assert len(train_indices) + len(test_indices) == datasize

    test_groups = groups[test_indices]
    train_groups = groups[train_indices]

    assert set(test_groups).intersection(set(train_groups)) == set()


def test_stratified_train_test_partition():
    # perhaps use hypothesis for fuzzing the data
    datasize = 10_000

    y = np.random.normal(0, 1, size=datasize)

    train_indices, valid_indices, test_indices = stratified_train_test_partition(
        np.arange(datasize), y, train_size=0.5, valid_size=0.25, test_size=0.25, shuffle=shuffle
    )

    assert len(train_indices) + len(test_indices) + len(valid_indices) == datasize

    # make sure we have no index overlap
    assert len(np.intersect1d(train_indices, valid_indices)) == 0
    assert len(np.intersect1d(train_indices, test_indices)) == 0

    # since we stratify we should have comparable medians
    train_mean = np.median(y[train_indices])
    valid_mean = np.median(y[valid_indices])
    test_mean = np.median(y[test_indices])
    assert (
        pytest.approx(train_mean, abs=0.1)
        == pytest.approx(valid_mean, abs=0.1)
        == pytest.approx(test_mean, abs=0.1)
    )

    # now, no validation set
    train_indices, _, test_indices = stratified_train_test_partition(
        np.arange(datasize), y, train_size=0.8, valid_size=0.0, test_size=0.2, shuffle=shuffle
    )

    assert len(train_indices) + len(test_indices) == datasize

    assert len(np.intersect1d(train_indices, test_indices)) == 0


@pytest.mark.parametrize("number_of_groups", [30, 50, 80, 500, 8_000])
@pytest.mark.parametrize("shuffle", [True, False])
def test_grouped_train_valid_test_partition(number_of_groups, shuffle):
    # we might also want to use grouping without stratification to test
    # extrapolation
    datasize = 10_000
    groups = np.random.choice(np.arange(number_of_groups), size=datasize)

    train_indices, valid_indices, test_indices = grouped_train_valid_test_partition(
        groups, train_size=0.5, valid_size=0.25, test_size=0.25, shuffle=shuffle
    )

    assert len(train_indices) + len(test_indices) + len(valid_indices) == datasize
    test_groups = groups[test_indices]
    train_groups = groups[train_indices]
    valid_groups = groups[valid_indices]
    assert set(test_groups).intersection(set(train_groups)) == set()
    assert set(test_groups).intersection(set(valid_groups)) == set()
    assert set(train_groups).intersection(set(valid_groups)) == set()

    train_indices, valid_indices, test_indices = grouped_train_valid_test_partition(
        groups, train_size=0.5, valid_size=0, test_size=0.5, shuffle=shuffle
    )

    assert len(train_indices) + len(test_indices) == datasize
    test_groups = groups[test_indices]
    train_groups = groups[train_indices]

    assert set(test_groups).intersection(set(train_groups)) == set()


@pytest.mark.parametrize("shuffle", [True])
def test_grouped_train_valid_test_partition_string_groups(shuffle):
    datasize = 10_000
    groups = np.random.choice(["a", "b", "c", "d", "e", "f"], size=datasize)
    train_indices, valid_indices, test_indices = grouped_train_valid_test_partition(
        groups, train_size=0.5, valid_size=0.25, test_size=0.25, shuffle=shuffle
    )

    assert len(train_indices) + len(test_indices) + len(valid_indices) == datasize
    test_groups = groups[test_indices]
    train_groups = groups[train_indices]
    valid_groups = groups[valid_indices]
    assert set(test_groups).intersection(set(train_groups)) == set()
    assert set(test_groups).intersection(set(valid_groups)) == set()
    assert set(train_groups).intersection(set(valid_groups)) == set()

    train_indices, valid_indices, test_indices = grouped_train_valid_test_partition(
        groups, train_size=0.5, valid_size=0, test_size=0.5, shuffle=shuffle
    )

    assert len(train_indices) + len(test_indices) == datasize
    test_groups = groups[test_indices]
    train_groups = groups[train_indices]

    assert set(test_groups).intersection(set(train_groups)) == set()


def test_get_train_valid_test_sizes():
    train_size, valid_size, test_size = get_train_valid_test_sizes(100, 0.5, 0.25, 0.25)
    assert train_size + valid_size + test_size == 100
    assert train_size + valid_size == 75


def test_check_fraction():
    with pytest.raises(ValueError):
        check_fraction(-1, 0, 0)

    with pytest.raises(ValueError):
        check_fraction(1.1, 1.0, 0)

    with pytest.raises(ValueError):
        check_fraction(0.9, 0.2, 0.1)

    with pytest.raises(ValueError):
        check_fraction(0.1, 0.2, 0.4)

    _ = check_fraction(0.8, 0.1, 0.1)


def test_downsample_splits():
    res = downsample_splits([np.arange(100), np.arange(100)], sample_frac=0.8)

    assert len(res[0]) == 80
    assert len(res[1]) == 80



# -*- coding: utf-8 -*-
"""Helper functions to submit RASPA simulations."""

import os
import subprocess
from pathlib import Path
from shutil import which
from tempfile import TemporaryDirectory
from typing import Callable, Union

from loguru import logger
from pymatgen.core import IStructure, Structure

from mofdscribe.featurizers.utils.eqeq import get_eqeq_charges

from .ff_builder import ff_builder
from ..tempdir import TEMPDIR

RUN_SCRIPT = """#! /bin/sh -f
export DYLD_LIBRARY_PATH=RASPA_DIR/lib
export LD_LIBRARY_PATH=RASPA_DIR/lib
RASPA_DIR/bin/simulate
"""


def detect_raspa_dir():
    """Detect the RASPA directory."""
    raspa_dir = which("simulate")
    if raspa_dir is None:
        raise ValueError("RASPA not found.")
    else:
        logger.info(
            "RASPA_DIR not set in environment and input." " Attempting automatic detection."
        )
        p = Path(which("simulate"))
        return os.path.join(*p.parts[:-2])


def call_eqeq(structure: Union[Structure, IStructure], filename: Union[str, os.PathLike]) -> None:
    """Call EqEq to get a structure file `filename` with the charges."""
    if isinstance(structure, Structure):
        structure = IStructure.from_sites(structure)
    s, _ = get_eqeq_charges(structure)
    # This is a weird hack because recent versions of ASE changed how they write
    # symmetry in CIFs. One representative issue is https://github.com/lsmo-epfl/curated-cofs-submission/issues/17
    s = s.replace("_space_group_name_H-M_alt", "_symmetry_space_group_name_H-M")
    with open(filename, "w") as handle:
        handle.write(s)


def run_raspa(
    structure: Union[Structure, IStructure],
    raspa_dir: Union[str, os.PathLike],
    simulation_script: str,
    ff_params: dict,
    parser: Callable,
    run_eqeq: bool = False,
):
    """Submit a simulation to RASPA.

    Args:
        structure (Union[Structure, IStructure]): Input structure for the framework.
        raspa_dir (Union[str, os.PathLike]): Used for the `RASPA_DIR` environment variable.
        simulation_script (str): RASPA input file.
        ff_params (dict): settings for the force field builder.
        parser (Callable): function that takes the simulation directory as input
            and returns the output.
        run_eqeq (bool): If true, runs eqeq before submitting the RASPA simulations.
            Defaults to False.

    Raises:
        ValueError: In case the simulation fails.

    Returns:
        output created by the parser function.
    """
    ff_results = ff_builder(ff_params)
    with TemporaryDirectory(dir=TEMPDIR) as tempdir:
        for k, v in ff_results.items():
            with open(
                os.path.join(tempdir, k.replace("_def", ".def").replace("molecule_", "")), "w"
            ) as handle:
                handle.write(v)

        with open(os.path.join(tempdir, "simulation.input"), "w") as handle:
            handle.write(simulation_script)

        with open(os.path.join(tempdir, "run.sh"), "w") as handle:
            run_template = RUN_SCRIPT.replace("RASPA_DIR", raspa_dir)
            handle.write(run_template)

        structure.to(filename=os.path.join(tempdir, "input.cif"), fmt="cif")
        if run_eqeq:
            try:
                call_eqeq(structure, os.path.join(tempdir, "input.cif"))
            except Exception as e:
                raise ValueError(f"Error running EqEq. Output: {e}")

        try:
            _ = subprocess.run(  # noqa: S607, nosec
                ["sh", "run.sh"],
                universal_newlines=True,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.PIPE,
                check=True,
                cwd=tempdir,
            )
        except subprocess.CalledProcessError as e:
            raise ValueError(f"Error running RASPA. Output: {e.output}  stderr: {e.stderr}")
        results = parser(os.path.join(tempdir))

    return results



# -*- coding: utf-8 -*-
import pytest
from sklearn.ensemble import RandomForestRegressor

from mofdscribe.bench.df_model import DFModel
from mofdscribe.bench.logkHCO2 import LogkHCO2OODBench
from mofdscribe.bench.mofbench import BenchResult
from mofdscribe.datasets import CoREDataset


@pytest.mark.xdist_group(name="core-ds")
def test_df_model():
    ds = CoREDataset()
    model = RandomForestRegressor(n_estimators=100)
    df_model = DFModel(model, features=list(ds.available_features))
    bench = LogkHCO2OODBench(df_model, name="bla", debug=True, patch_in_ds=True)
    res = bench.bench()
    assert isinstance(res, BenchResult)



# -*- coding: utf-8 -*-
"""Make sure that the utils work."""
import pytest
from matminer.featurizers.base import BaseFeaturizer
from pymatgen.core import IMolecule, IStructure, Molecule, Structure

from mofdscribe.featurizers.utils.extend import (
    operates_on_imolecule,
    operates_on_istructure,
    operates_on_molecule,
    operates_on_structure,
)


class TestFeaturizer(BaseFeaturizer):
    def citations(self):
        return []

    def featurize(self, structure):
        return [0]

    def feature_labels(self):
        return ["test"]

    def implementors(self):
        return []


def test_add_operates_on():
    """Test that the add_operates_on decorator works."""

    class TestClass:
        pass

    decorator_test = operates_on_structure(TestClass)

    assert decorator_test().operates_on() == [Structure]

    with pytest.raises(AttributeError):
        TestFeaturizer()._accepted_types

    with pytest.raises(AttributeError):
        TestFeaturizer().operates_on()

    @operates_on_molecule
    class DecoratorTest2(TestFeaturizer):
        pass

    assert DecoratorTest2().operates_on() == [Molecule]

    @operates_on_structure
    @operates_on_molecule
    @operates_on_imolecule
    @operates_on_istructure
    class DecoratorTest3(TestFeaturizer):
        pass

    assert set(DecoratorTest3().operates_on()) == {Structure, Molecule, IMolecule, IStructure}



# -*- coding: utf-8 -*-
"""Metrics for the regression setting."""
from __future__ import annotations

from typing import List

import numpy as np
from numpy.typing import ArrayLike
from pydantic import BaseModel
from sklearn.metrics import (
    max_error,
    mean_absolute_error,
    mean_absolute_percentage_error,
    mean_squared_error,
    r2_score,
)

__all__ = ["RegressionMetrics", "get_regression_metrics", "top_n_in_top_k"]


def top_n_in_top_k(
    predictions: ArrayLike, labels: ArrayLike, k: int, n: int, maximize: bool = True
) -> int:
    """Find how many of the top n predictions are in the top k labels.

    Args:
        predictions (ArrayLike): predictions for one objective
        labels (ArrayLike): true labels for one objective
        k (int): number of top labels to consider
        n (int): number of top predictions to consider
        maximize (bool): Set to `True` if larger is better.
            Defaults to True.

    Examples:
        >>> predictions = [0.1, 0.2, 0.3, 0.4, 0.5]
        >>> labels = [0.1, 0.2, 0.3, 0.4, 0.5]
        >>> top_n_in_top_k(predictions, labels, k=2, n=2)
        2

    Returns:
        int: number of top n predictions in top k labels
    """
    indices_predictions = np.argsort(predictions)
    indices_labels = np.argsort(labels)

    if maximize:
        indices_predictions = indices_predictions[::-1]
        indices_labels = indices_labels[::-1]

    top_n_predictions = indices_predictions[:n]
    top_k_labels = indices_labels[:k]

    return np.sum(np.isin(top_n_predictions, top_k_labels))


class RegressionMetrics(BaseModel):
    """Model for regression metrics.

    We will use this model to validate the benchmark results.
    """

    mean_squared_error: float
    mean_absolute_error: float
    r2_score: float
    max_error: float
    mean_absolute_percentage_error: float
    top_5_in_top_5: int
    top_10_in_top_10: int
    top_50_in_top_50: int
    top_100_in_top_100: int
    top_500_in_top_500: int


class RegressionMetricsConcat(BaseModel):
    mean_squared_error: List[float]
    mean_absolute_error: List[float]
    r2_score: List[float]
    max_error: List[float]
    mean_absolute_percentage_error: List[float]
    top_5_in_top_5: List[int]
    top_10_in_top_10: List[int]
    top_50_in_top_50: List[int]
    top_100_in_top_100: List[int]
    top_500_in_top_500: List[int]


def get_regression_metrics(predictions: ArrayLike, labels: ArrayLike) -> RegressionMetrics:
    """Get regression metrics.

    Args:
        predictions (ArrayLike): predictions for one objective
        labels (ArrayLike): true labels for one objective

    Returns:
        RegressionMetrics: regression metrics

    Examples:
        >>> predictions = [0.1, 0.2, 0.3, 0.4, 0.5]
        >>> labels = [0.1, 0.2, 0.3, 0.4, 0.5]
        >>> get_regression_metrics(predictions, labels)
        RegressionMetrics(**{'mean_absolute_error': 0.0,
        'mape': 0.0,
        'mean_squared_error': 0.0,
        'r2': 1.0,
        'max_error': 0.0,
        'top_5_in_top_5': 5,
        'top_10_in_top_10': 10,
        'top_100_in_top_100': 100,
        'top_500_in_top_500': 500})
    """
    metrics = RegressionMetrics(
        **{
            "mean_squared_error": mean_squared_error(labels, predictions),
            "mean_absolute_error": mean_absolute_error(labels, predictions),
            "r2_score": r2_score(labels, predictions),
            "max_error": max_error(labels, predictions),
            "mean_absolute_percentage_error": mean_absolute_percentage_error(labels, predictions),
            "top_5_in_top_5": top_n_in_top_k(predictions, labels, k=5, n=5),
            "top_10_in_top_10": top_n_in_top_k(predictions, labels, k=10, n=10),
            "top_50_in_top_50": top_n_in_top_k(predictions, labels, k=50, n=50),
            "top_100_in_top_100": top_n_in_top_k(predictions, labels, k=100, n=100),
            "top_500_in_top_500": top_n_in_top_k(predictions, labels, k=500, n=500),
        }
    )

    return metrics



# -*- coding: utf-8 -*-
"""Test that we can convert pymatgen Molecules to RDkit molecules."""
from mofdscribe.featurizers.bu.utils import create_rdkit_mol_from_mol_graph


def test_conversion_to_rdkit(molecule_graph):
    """Test that we can convert pymatgen Molecules to RDkit molecules."""
    rdkit_mol = create_rdkit_mol_from_mol_graph(molecule_graph)
    assert rdkit_mol.GetNumAtoms() == molecule_graph.graph.number_of_nodes()



# -*- coding: utf-8 -*-
"""Extract substructures (e.g. of certain element types)."""
from typing import Collection, List, Union

from pymatgen.core import IMolecule, IStructure, Molecule, Structure


def filter_element(
    structure: Union[Structure, IStructure, Molecule, IMolecule], elements: List[str]
) -> Structure:
    """Filter a structure by element.

    Args:
        structure (Union[Structure, IStructure, Molecule, IMolecule]): input structure
        elements (str): element to filter

    Returns:
        filtered_structure (Structure): filtered structure
    """
    elements_ = []
    for atom_type in elements:
        if "-" in atom_type:
            elements_.extend(atom_type.split("-"))
        else:
            elements_.append(atom_type)
    keep_sites = []
    for site in structure.sites:
        if site.specie.symbol in elements_:
            keep_sites.append(site)
    if len(keep_sites) == 0:
        return None

    input_is_structure = isinstance(structure, (Structure, IStructure))
    if input_is_structure:
        return Structure.from_sites(keep_sites)
    else:  # input is molecule or IMolecule
        return Molecule.from_sites(keep_sites)


def elements_in_structure(structure: Structure) -> List[str]:
    """Get all elements symbols in a structure."""
    return list(structure.composition.get_el_amt_dict().keys())


def get_metal_indices(structure: Structure) -> List[int]:
    """Get the indices of metals in the structure."""
    metal_indices = []
    for i, s in enumerate(structure):
        if s.specie.is_metal:
            metal_indices.append(i)
    return metal_indices


def select_elements(structure: Structure, element: str) -> Structure:
    """Return a structure with only the given element."""
    sites = []
    for site in structure:
        if str(site.specie) == element:
            sites.append(site)

    return Structure.from_sites(sites)


def _not_relevant_structure_indices(structure: Union[Structure, IStructure], indices: Collection):
    """Return the indices of the sites that are not included in the provided indices."""
    return [i for i in range(len(structure)) if i not in indices]



# -*- coding: utf-8 -*-
"""Code taken from the SI for 10.1021/acs.jcim.6b00565"""
from collections import OrderedDict
from typing import List

import numpy as np
from rdkit import Chem
from rdkit.Chem import AllChem

from .rdkitadaptor import RDKitAdaptor


# ToDo: Allow to set some of the options, e.g. for UFF
def _generate_conformers(mol, num_confs):
    # Add H atoms to skeleton
    molecule = Chem.AddHs(mol)
    conformer_integers = []
    # Embed and optimise the conformers
    conformers = AllChem.EmbedMultipleConfs(molecule, num_confs, pruneRmsThresh=0.5, numThreads=5)

    optimised_and_energies = AllChem.MMFFOptimizeMoleculeConfs(
        molecule, maxIters=600, numThreads=5, nonBondedThresh=100.0
    )
    energy_dict_with_key_was_id = {}
    final_conformers_to_use = {}
    # Only keep the conformers which were successfully fully optimised

    for conformer in conformers:
        optimised, energy = optimised_and_energies[conformer]
        if optimised == 0:
            energy_dict_with_key_was_id[conformer] = energy
            conformer_integers.append(conformer)
    # Keep the lowest energy conformer
    lowestenergy = min(energy_dict_with_key_was_id.values())
    for key, value in energy_dict_with_key_was_id.items():
        if value == lowestenergy:
            lowest_energy_conformer_id = key
    final_conformers_to_use[lowest_energy_conformer_id] = lowestenergy

    # Remove H atoms to speed up substructure matching
    molecule = AllChem.RemoveHs(molecule)
    # Find all substructure matches of the molecule with itself,
    # to account for symmetry
    matches = molecule.GetSubstructMatches(molecule, uniquify=False)
    maps = [list(enumerate(match)) for match in matches]
    # Loop over conformers other than the lowest energy one
    for conformer_id, _ in energy_dict_with_key_was_id.items():
        okay_to_add = True
        for finalconformer_id in final_conformers_to_use:
            rms = AllChem.GetBestRMS(molecule, molecule, finalconformer_id, conformer_id, maps)
            if rms < 1.0:
                okay_to_add = False
                break

        if okay_to_add:
            final_conformers_to_use[conformer_id] = energy_dict_with_key_was_id[conformer_id]

    sorted_dictionary = OrderedDict(sorted(final_conformers_to_use.items(), key=lambda t: t[1]))
    energies = list(sorted_dictionary.values())

    return energies


def _calc_nconf20(energy_list):
    energy_descriptor = 0

    relative_energies = np.array(energy_list) - energy_list[0]

    for energy in relative_energies[1:]:
        if 0 <= energy < 20:
            energy_descriptor += 1

    return energy_descriptor


def _n_conf20(mol):
    try:
        energy_list = _generate_conformers(mol, 100)
        descriptor = _calc_nconf20(energy_list)
        return np.array([descriptor])
    except Exception:
        return np.array([np.nan])


class NConf20(RDKitAdaptor):
    """
    Compute the nConf20 descriptor for a molecule.

    This descriptor attempts to capture the flexibility of molecules
    by sampling the conformational space. The descriptor is a count
    of the "accessible" conformers (based on relative conformer energies up to 20 kcal/mol,
    the lowest energy conformer is not counted).
    Conformers are generated using the RDKit conformer generator.

    ... warning::
        Part of the featurization is a geometry optimization using the MMFF
        force field. This will naturally fail for some molecules, and for all metal clusters.
        In these cases, the descriptor will be set to NaN.
    """

    def __init__(self):
        """Construct a new nConf20 featurizer."""
        super().__init__(featurizer=_n_conf20, feature_labels=["n_conf20"])

    def citations(self) -> List[str]:
        return super().citations() + [
            "@article{Wicker2016,"
            "doi = {10.1021/acs.jcim.6b00565},"
            "url = {https://doi.org/10.1021/acs.jcim.6b00565},"
            "year = {2016},"
            "month = dec,"
            "publisher = {American Chemical Society ({ACS})},"
            "volume = {56},"
            "number = {12},"
            "pages = {2347--2352},"
            "author = {Jerome G. P. Wicker and Richard I. Cooper},"
            "title = {Beyond Rotatable Bond Counts: Capturing 3D Conformational Flexibility in a Single Descriptor},"
            "journal = {Journal of Chemical Information and Modeling}"
            "}",
            "@article{tosco2014bringing,"
            "title={Bringing the MMFF force field to the RDKit: implementation and validation},"
            "author={Tosco, Paolo and Stiefl, Nikolaus and Landrum, Gregory},"
            "journal={Journal of cheminformatics},"
            "volume={6},"
            "number={1},"
            "pages={1--4},"
            "year={2014},"
            "publisher={Springer}"
            "}",
        ]

    def implementors(self) -> List[str]:
        return super().implementors() + ["Jerome G. P. Wicker and Richard I. Cooper"]



# -*- coding: utf-8 -*-
"""Code based on the pyntcloud library.

The MIT License

Copyright (c) 2017-2019 The pyntcloud Developers

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from typing import List

import numpy as np
from scipy.spatial import cKDTree


def cartesian(arrays: List[np.typing.ArrayLike], out: np.ndarray = None) -> np.ndarray:
    """Generate a Cartesian product of input arrays.

    Args:
        arrays (List[np.typing.ArrayLike]): list of array-like 1-D arrays to form the
            Cartesian product of.
        out (np.ndarray): Array to  place the cartesian product in.
            Defaults to None.

    Returns:
        np.ndarray: 2-D array of shape (M, len(arrays)) containing cartesian
        products formed of input arrays.

    Examples:
        >>> cartesian(([1, 2, 3], [4, 5], [6, 7]))
        array([[1, 4, 6],
            [1, 4, 7],
            [1, 5, 6],
            [1, 5, 7],
            [2, 4, 6],
            [2, 4, 7],
            [2, 5, 6],
            [2, 5, 7],
            [3, 4, 6],
            [3, 4, 7],
            [3, 5, 6],
            [3, 5, 7]])
    """
    arrays = [np.asarray(x) for x in arrays]
    shape = (len(x) for x in arrays)
    dtype = arrays[0].dtype

    ix = np.indices(shape)
    ix = ix.reshape(len(arrays), -1).T

    if out is None:
        out = np.empty_like(ix, dtype=dtype)

    for n, _ in enumerate(arrays):
        out[:, n] = arrays[n][ix[:, n]]

    return out


class VoxelGrid:
    def __init__(
        self,
        points,
        properties=None,
        n_x=1,
        n_y=1,
        n_z=1,
        size_x=None,
        size_y=None,
        size_z=None,
        regular_bounding_box=True,
    ):
        """Initialize a VoxelGrid.

        Args:
            points (np.ndarray): shape (N,  3)
            properties (numpy.array): Shape (N, 3). Defaults to None.
            n_x (int): The number of segments in which each axis will be divided.
                Ignored if corresponding size_x, size_y or size_z is not None. Defaults to 1.
            n_y (int): The number of segments in which each axis will be divided.
                Ignored if corresponding size_x, size_y or size_z is not None. Defaults to 1.
            n_z (int):  The number of segments in which each axis will be divided.
                Ignored if corresponding size_x, size_y or size_z is not None. Defaults to 1.
            size_x (float): The desired voxel size along each axis.
                If not None, the corresponding n_x, n_y or n_z will be ignored.
                Defaults to None.
            size_y (float): The desired voxel size along each axis.
                If not None, the corresponding n_x, n_y or n_z will be ignored.
                Defaults to None.
            size_z (float): The desired voxel size along each axis.
                If not None, the corresponding n_x, n_y or n_z will be ignored.
                Defaults to None.
            regular_bounding_box (bool): If True, the bounding box
                of the point cloud will be adjusted in order to have all
                the dimensions of equal length. Defaults to True.
        """
        self._points = points
        self.properties = properties
        self.x_y_z = np.asarray([n_x, n_y, n_z])
        self.sizes = np.asarray([size_x, size_y, size_z])
        self.regular_bounding_box = regular_bounding_box

        self.id = None
        self.xyzmin, self.xyzmax = None, None
        self.segments = None
        self.shape = None
        self.n_voxels = None
        self.voxel_x, self.voxel_y, self.voxel_z = None, None, None
        self.voxel_n = None
        self.voxel_centers = None
        self.voxel_colors = None

    def compute(self):
        """Compute the voxel grid."""
        xyzmin = self._points.min(0)
        xyzmax = self._points.max(0)
        xyz_range = self._points.ptp(0)

        if self.regular_bounding_box:
            #: adjust to obtain a minimum bounding box with all sides of equal length
            margin = max(xyz_range) - xyz_range
            xyzmin = xyzmin - margin / 2
            xyzmax = xyzmax + margin / 2

        for n, size in enumerate(self.sizes):
            if size is None:
                continue
            margin = (((self._points.ptp(0)[n] // size) + 1) * size) - self._points.ptp(0)[n]
            xyzmin[n] -= margin / 2
            xyzmax[n] += margin / 2
            self.x_y_z[n] = ((xyzmax[n] - xyzmin[n]) / size).astype(int)

        self.xyzmin = xyzmin
        self.xyzmax = xyzmax

        segments = []
        shape = []
        for i in range(3):
            # note the +1 in num
            s, step = np.linspace(xyzmin[i], xyzmax[i], num=(self.x_y_z[i] + 1), retstep=True)
            segments.append(s)
            shape.append(step)

        self.segments = segments
        self.shape = shape

        self.n_voxels = np.prod(self.x_y_z)

        self.id = "V({},{},{})".format(self.x_y_z, self.sizes, self.regular_bounding_box)

        # find where each point lies in corresponding segmented axis
        # -1 so index are 0-based; clip for edge cases
        self.voxel_x = np.clip(
            np.searchsorted(self.segments[0], self._points[:, 0]) - 1, 0, self.x_y_z[0]
        )
        self.voxel_y = np.clip(
            np.searchsorted(self.segments[1], self._points[:, 1]) - 1, 0, self.x_y_z[1]
        )
        self.voxel_z = np.clip(
            np.searchsorted(self.segments[2], self._points[:, 2]) - 1, 0, self.x_y_z[2]
        )
        # for each point get the index of the voxel it belongs to
        self.voxel_n = np.ravel_multi_index([self.voxel_x, self.voxel_y, self.voxel_z], self.x_y_z)

        # compute center of each voxel
        midsegments = [(self.segments[i][1:] + self.segments[i][:-1]) / 2 for i in range(3)]
        self.voxel_centers = cartesian(midsegments).astype(np.float32)

        if self.properties is not None:
            order = np.argsort(self.voxel_n)

            _, breaks, counts = np.unique(
                self.voxel_n[order], return_index=True, return_counts=True
            )
            repeated_counts = np.repeat(counts[:, None], self.properties.shape[1], axis=1)
            summed_colors = np.add.reduceat(self.properties, breaks, axis=0)
            averaged_colors = summed_colors / repeated_counts
            self.averaged_properties = averaged_colors

    def query(self, points):
        """Query structure."""
        voxel_x = np.clip(np.searchsorted(self.segments[0], points[:, 0]) - 1, 0, self.x_y_z[0])
        voxel_y = np.clip(np.searchsorted(self.segments[1], points[:, 1]) - 1, 0, self.x_y_z[1])
        voxel_z = np.clip(np.searchsorted(self.segments[2], points[:, 2]) - 1, 0, self.x_y_z[2])
        voxel_n = np.ravel_multi_index([voxel_x, voxel_y, voxel_z], self.x_y_z)

        return voxel_n

    def get_feature_vector(self, mode="binary", flatten: bool = False):
        """Get feature vector.

        Args:
            mode (str): Available modes are:
                * binary: 0 for empty voxels, 1 for occupied.
                * density: number of points inside voxel / total number of points.
                * TDF: Truncated Distance Function. Value between 0 and 1 indicating
                the distance between the voxel's center and the closest point. 1 on the surface,
                0 on voxels further than 2 * voxel side.
                Defaults to "binary".
            flatten (bool): Returns a flattened vector.
                Defaults to False.

        Raises:
            NotImplementedError: If the mode is not implemented.

        Returns:
            np.ndarray: _description_
        """
        vector = np.zeros(self.n_voxels)

        if mode == "binary":
            vector[np.unique(self.voxel_n)] = 1

        elif mode == "density":
            count = np.bincount(self.voxel_n)
            vector[: len(count)] = count
            vector /= len(self.voxel_n)

        elif mode == "TDF":
            kdt = cKDTree(self._points)
            vector, i = kdt.query(self.voxel_centers, n_jobs=-1)

        elif isinstance(mode, int):
            unique_voxels = np.unique(self.voxel_n)
            vector[unique_voxels] = self.averaged_properties[:, mode]

        else:
            raise NotImplementedError("{} is not a supported feature vector mode".format(mode))

        if flatten:
            return vector
        return vector.reshape(self.x_y_z)

    def get_voxel_neighbors(self, voxel):
        """Get all voxel neighbors.

        Args:
            voxel (int): Voxel index.

        Returns:
            List[int]: Indices of the valid, non-empty 26 neighborhood around voxel.
        """
        x, y, z = np.unravel_index(voxel, self.x_y_z)

        valid_x = []
        valid_y = []
        valid_z = []
        if x - 1 >= 0:
            valid_x.append(x - 1)
        if y - 1 >= 0:
            valid_y.append(y - 1)
        if z - 1 >= 0:
            valid_z.append(z - 1)

        valid_x.append(x)
        valid_y.append(y)
        valid_z.append(z)

        if x + 1 < self.x_y_z[0]:
            valid_x.append(x + 1)
        if y + 1 < self.x_y_z[1]:
            valid_y.append(y + 1)
        if z + 1 < self.x_y_z[2]:
            valid_z.append(z + 1)

        valid_neighbor_indices = cartesian((valid_x, valid_y, valid_z))

        ravel_indices = np.ravel_multi_index(
            (
                valid_neighbor_indices[:, 0],
                valid_neighbor_indices[:, 1],
                valid_neighbor_indices[:, 2],
            ),
            self.x_y_z,
        )

        return [x for x in ravel_indices if x in np.unique(self.voxel_n)]



# -*- coding: utf-8 -*-
"""Test the RDkitAdaptor."""
import numpy as np
from rdkit.Chem.Descriptors3D import Asphericity

from mofdscribe.featurizers.bu.rdkitadaptor import RDKitAdaptor


def test_rdkit_adaptor(molecule_graph, molecule):
    """Test that we can call RDKit featurizers with pymatgen molecules."""
    adaptor = RDKitAdaptor(Asphericity, ["asphericity"], "vesta")

    features_a = adaptor.featurize(molecule)
    assert features_a.shape == (1,)
    assert features_a.dtype == np.float

    features_b = adaptor.featurize(molecule_graph)
    assert all(features_a == features_b)



# -*- coding: utf-8 -*-
"""Describe molecules by computing a histogram of pairwise distances between their atoms."""
from typing import List, Union

import numpy as np
from matminer.featurizers.base import BaseFeaturizer
from pymatgen.core import IMolecule, IStructure, Molecule, Structure

from mofdscribe.featurizers.utils.extend import (
    operates_on_imolecule,
    operates_on_istructure,
    operates_on_molecule,
    operates_on_structure,
)
from mofdscribe.featurizers.utils.histogram import get_rdf


@operates_on_molecule
@operates_on_imolecule
@operates_on_istructure
@operates_on_structure
class PairwiseDistanceHist(BaseFeaturizer):
    """
    Describe the shape of molecules by computing a histogram of pairwise distances.

    For doing so, we will just compute all pairwise distances and then compute
    the histogram of them.
    One might also think of this as pretty rough approximation of something like
    the AMD fingerpint

    It also has some similarities to the "Grouped representation of interatomic distances"
    reported in [Zhang2022]_.
    """

    def __init__(
        self,
        lower_bound: float = 0.0,
        upper_bound: float = 15.0,
        bin_size: float = 0.5,
        density: bool = True,
    ) -> None:
        """Create a new PairwiseDistanceHist featurizer.

        Args:
            lower_bound (float): Lower bound of the histogram.
                Defaults to 0.0.
            upper_bound (float): Upper bound of the histogram.
                Defaults to 15.0.
            bin_size (float): Size of the bins.
                Defaults to 0.5.
            density (bool): Whether to return the density or the counts.
                Defaults to True.
        """
        self.lower_bound = lower_bound
        self.upper_bound = upper_bound
        self.bin_size = bin_size
        self.density = density

    def _get_grid(self):
        return np.arange(self.lower_bound, self.upper_bound, self.bin_size)

    def feature_labels(self) -> List[str]:
        return [f"pairwise_distance_hist_{a}" for a in self._get_grid()]

    def featurize(self, structure: Union[Molecule, IMolecule, Structure, IStructure]) -> np.ndarray:
        distances = []
        for i, _ in enumerate(structure):
            for j, _ in enumerate(structure):
                if i < j:
                    distances.append(structure.get_distance(i, j))

        features = get_rdf(
            distances,
            lower_lim=self.lower_bound,
            upper_lim=self.upper_bound,
            bin_size=self.bin_size,
            density=self.density,
            normalized=False,
        )
        return features

    def implementors(self) -> List[str]:
        return ["Kevin Maik Jablonka"]

    def citations(self) -> List[str]:
        return []



# -*- coding: utf-8 -*-
"""Helper functions for the splitters.

Some of these methods might also be useful for constructing nested cross-validation loops.
"""
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd
from loguru import logger
from numpy.typing import ArrayLike
from pandas.api.types import infer_dtype
from scipy.spatial.distance import cdist
from sklearn.cluster import KMeans
from sklearn.decomposition import PCA
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import StandardScaler
from tqdm import tqdm


# todo: can we do here something with numba?
# for numba, we would need to do some rewrite as there is no support
# for numpy.delete
def kennard_stone_sampling(
    X: ArrayLike,  # noqa: N803
    scale: bool = True,
    centrality_measure: str = "mean",
    metric: Union[Callable, str] = "euclidean",
) -> List[int]:
    """Run the Kennard-Stone sampling algorithm [KennardStone].

    The algorithm selects samples with uniform converage.
    The initial samples are biased towards the boundaries of the dataset.

    .. note::

        You also might this algorithm useful for creating a "diverse"
        sample of points, e.g., to initalize an active learnign loop.

    .. warning::

        This algorithm has a high computational complexity.
        It is not recommended for large datasets.

    Args:
        X (ArrayLike): Input feature matrix.
        scale (bool): If True, apply z-score normalization
            prior to running the sampling. Defaults to True.
        centrality_measure (str): The first sample is selected to be
            maximally distanct from this value. It can be one of "mean", "median",
            "random". In case of "random" we simply select a random point.
            In the case of "mean" and "median" the initial point is maximally distanct
            from the mean and median of the feature matrix, respectively.
            Defaults to "mean".
        metric (Union[Callable, str]): The distance metric to use.
            If a string, the distance function can be ‘braycurtis’, ‘canberra’, ‘chebyshev’,
            ‘cityblock’, ‘correlation’, ‘cosine’, ‘dice’, ‘euclidean’, ‘hamming’, ‘jaccard’,
            ‘jensenshannon’, ‘kulsinski’, ‘kulczynski1’, ‘mahalanobis’, ‘matching’, ‘minkowski’,
            ‘rogerstanimoto’, ‘russellrao’, ‘seuclidean’, ‘sokalmichener’, ‘sokalsneath’, ‘sqeuclidean’,
            ‘yule’. See :py:meth:`scipy.spatial.distance.cdist`.

            Defaults to "euclidean".

    Raises:
        ValueError: If non-implemented centrality measure is used.

    Returns:
        List[int]: indices sorted by their Max-Min distance.
    """
    if scale:
        scaler = StandardScaler()
        X = scaler.fit_transform(X)  # noqa: N806

    original_X = X.copy()  # noqa: N806

    if centrality_measure == "mean":
        distance_to_center = cdist(X, X.mean(axis=0).reshape(1, -1), metric=metric)
    elif centrality_measure == "median":
        distance_to_center = cdist(X, np.median(X, axis=0).reshape(1, -1), metric=metric)
    elif centrality_measure == "random":
        distance_to_center = cdist(X, X[np.random.choice(X.shape[0])].reshape(1, -1), metric=metric)
    else:
        raise ValueError(f"Unknown centrality measure: {centrality_measure}")

    index_farthest = np.argmax(distance_to_center)

    index_selected = [index_farthest]

    index_remaining = np.arange(len(X))

    X = np.delete(X, index_selected, axis=0)  # noqa: N806
    index_remaining = np.delete(index_remaining, index_selected, axis=0)

    # we will sometimes need to do quite a few steps.
    # for all practical datasets this will exceeed python's recursion limit.
    # we therefore use a while loop to avoid this.
    with tqdm(total=len(X) - 1) as pbar:
        while len(index_remaining):
            samples_selected = original_X[index_selected]
            min_distance_to_samples_selected = np.min(cdist(samples_selected, X, metric=metric))

            index_farthest = np.argmax(min_distance_to_samples_selected)

            index_selected.append(index_remaining[index_farthest])
            X = np.delete(X, index_farthest, axis=0)  # noqa: N806
            index_remaining = np.delete(index_remaining, index_farthest, 0)
            pbar.update(1)

    return index_selected


def pca_kmeans(
    X: np.ndarray,  # noqa: N803
    scaled: bool,
    n_pca_components: Union[int, str],
    n_clusters: int,
    random_state: Optional[Union[int, np.random.RandomState]] = None,
    pca_kwargs: Optional[Dict[str, Any]] = None,
    kmeans_kwargs: Optional[Dict[str, Any]] = None,
) -> np.ndarray:
    """Run principal component analysis (PCA) followed by K-means clustering on the data.

    Uses sklearn's implementation of PCA, and k-means.

    Args:
        X (np.ndarray): Input data
        scaled (bool): If True, use standard scaling for clustering
        n_pca_components (Union[int, str]): number of principal components to keep
        n_clusters (int): number of clusters
        random_state (Optional[Union[int, np.random.RandomState]], optional): Random state for sklearn.
            Defaults to None.
        pca_kwargs (Dict[str, Any], optional): Additional keyword arguments for
            sklearn's :py:class:`sklearn.decomposition.PCA`. Defaults to None.
        kmeans_kwargs (Dict[str, Any], optional):  Additional keyword arguments for
            sklearn's :py:class:`sklearn.clustering.KMeans`. Defaults to None.

    Returns:
        np.ndarray: Cluster indices.
    """
    if scaled:
        X = StandardScaler().fit_transform(X)  # noqa: N806

    if n_pca_components is not None:
        pca = PCA(n_components=n_pca_components, **(pca_kwargs or {}))
        X_pca = pca.fit_transform(X)  # noqa: N806
    else:
        X_pca = X  # noqa: N806
    kmeans = KMeans(n_clusters=n_clusters, random_state=random_state, **(kmeans_kwargs or {}))
    return kmeans.fit_predict(X_pca)


def is_categorical(x: Union[float, int, np.typing.ArrayLike]) -> bool:
    """Return true if x is categorial or composed of integers."""
    return infer_dtype(x) == "category" or infer_dtype(x) == "integer"


def stratified_train_test_partition(
    idxs: Sequence[int],
    stratification_col: np.typing.ArrayLike,
    train_size: float,
    valid_size: float,
    test_size: float,
    shuffle: bool = True,
    random_state: Optional[Union[int, np.random.RandomState]] = None,
    q: Sequence[float] = (0, 0.25, 0.5, 0.75, 1),
) -> Tuple[np.array, np.array, np.array]:
    """Perform a stratified train/test split.

    .. seealso::

        * :py:meth:`mofdscribe.splitters.utils.grouped_stratified_train_test_partition`:
          performs an grouped stratified train/test split
        * :py:meth:`mofdscribe.splitters.utils.grouped_train_valid_test_partition`:
          performs an grouped un-stratified train/test split

    Args:
        idxs (Sequence[int]): Indices of points to split
        stratification_col (np.typing.ArrayLike): Data used for stratification.
            If it is categorical (see :py:meth:`mofdscribe.splitters.utils.is_categorical`)
            then we directly use it for stratification. Otherwise, we use quantile binning.
        train_size (float): Size of the training set as fraction.
        valid_size (float): Size of the validation set as fraction.
        test_size (float): Size of the test set as fraction.
        shuffle (bool): If True, perform a shuffled split. Defaults to True.
        random_state (Union[int, np.random.RandomState], optional):
            Random state for the suffler. Defaults to None.
        q (Sequence[float], optional): List of quantiles used for quantile binning.
            Defaults to (0, 0.25, 0.5, 0.75, 1).

    Returns:
        Tuple[np.array, np.array, np.array]: Train, validation, test indices.
    """
    if stratification_col is not None:
        if is_categorical(stratification_col):
            stratification_col = stratification_col
        else:
            logger.warning(
                "Stratifying on non-categorical data. "
                "Note that there is still discussion on the usefullness of this method."
            )

            stratification_col = quantile_binning(stratification_col, q=q)

    train_size, valid_size, test_size = get_train_valid_test_sizes(
        len(stratification_col), train_size, valid_size, test_size
    )

    train_idx, test_idx, train_strat, _ = train_test_split(
        idxs,
        stratification_col,
        train_size=train_size + valid_size,
        test_size=test_size,
        shuffle=shuffle,
        random_state=random_state,
        stratify=stratification_col,
    )

    if valid_size > 0:
        train_idx, valid_idx = train_test_split(
            train_idx,
            train_size=train_size,
            shuffle=shuffle,
            random_state=random_state,
            stratify=train_strat,
        )
    else:
        valid_idx = []

    return np.array(train_idx), np.array(valid_idx), np.array(test_idx)


def grouped_stratified_train_test_partition(
    stratification_col: np.typing.ArrayLike,
    group_col: np.typing.ArrayLike,
    train_size: float,
    valid_size: float,
    test_size: float,
    shuffle: bool = True,
    random_state: Optional[Union[int, np.random.RandomState]] = None,
    q: Sequence[float] = (0, 0.25, 0.5, 0.75, 1),
    center: Callable = np.median,
) -> Tuple[np.array, np.array, np.array]:
    """Return grouped stratified train-test partition.

    First, we compute the most common stratification category / centrality measure
    of the stratification column for every group.
    Then, we perform a stratified train/test partition on the groups.
    We then "expand" by concatenating the indices belonging to each group.

    .. warning::

        Note that this won't work well if the number of groups and datapoints is small.
        It will also cause issues if the number of datapoints in the groups is very
        imbalanced.

    .. seealso::

        * :py:meth:`mofdscribe.splitters.utils.stratified_train_test_partition`:
          performs an un-grouped stratified train/test split
        * :py:meth:`mofdscribe.splitters.utils.grouped_train_valid_test_partition`:
          performs an grouped un-stratified train/test split

    Args:
        stratification_col (np.typing.ArrayLike):  Data used for stratification.
            If it is categorical (see :py:meth:`mofdscribe.splitters.utils.is_categorical`)
            then we directly use it for stratification. Otherwise, we use quantile binning.
        group_col (np.typing.ArrayLike): Data used for grouping.
        train_size (float): Size of the training set as fraction.
        valid_size (float): Size of the validation set as fraction.
        test_size (float): Size of the test set as fraction.
        shuffle (bool): If True, perform a shuffled split. Defaults to True.
        random_state (Union[int, np.random.RandomState], optional):
            Random state for the suffler. Defaults to None.
        q (Sequence[float], optional): List of quantiles used for quantile binning.
            Defaults to [0, 0.25, 0.5, 0.75, 1].
        center (Callable): Aggregation function to compute a measure of centrality
            of all the points in a group such that this can then be used for stratification.
            This is only used for continuos inputs. For categorical inputs, we always use
            the mode.

    Returns:
        Tuple[np.array, np.array, np.array]: Train, validation, test indices.
    """
    groups = np.unique(group_col)
    categorical = is_categorical(stratification_col)

    category_for_group = []
    # for each group, get the mode
    if categorical:
        for group in groups:
            category_for_group.append(np.mode(stratification_col[group_col == group]))
    else:
        for group in groups:
            category_for_group.append(center(stratification_col[group_col == group]))

    # if we do not have categories, we now need to discretize the stratification_col
    if not categorical:
        category_for_group = quantile_binning(category_for_group, q=q)

    train_size, valid_size, test_size = get_train_valid_test_sizes(
        len(category_for_group), train_size, valid_size, test_size
    )

    # now we can do the split
    train_groups, test_groups, train_cat, _ = train_test_split(
        groups,
        category_for_group,
        train_size=train_size + valid_size,
        test_size=test_size,
        stratify=category_for_group,
        shuffle=shuffle,
        random_state=random_state,
    )

    if valid_size > 0:
        train_groups, valid_groups = train_test_split(
            train_groups,
            train_size=train_size,
            stratify=train_cat,
            shuffle=shuffle,
            random_state=random_state,
        )

    # now, get the original indices
    train_indices = np.where(np.isin(group_col, train_groups))[0]
    if valid_size > 0:
        valid_indices = np.where(np.isin(group_col, valid_groups))[0]
    else:
        valid_indices = []
    test_indices = np.where(np.isin(group_col, test_groups))[0]

    return train_indices, valid_indices, test_indices


def get_train_valid_test_sizes(
    size: int, train_size: float, valid_size: float, test_size: float
) -> Tuple[int, int, int]:
    """Compute the number of points in every split."""
    train_size = int(np.floor(train_size * size))
    valid_size = int(np.ceil(valid_size * size))
    test_size = int(size - train_size - valid_size)
    return train_size, valid_size, test_size


def grouped_train_valid_test_partition(
    groups: np.typing.ArrayLike,
    train_size: float,
    valid_size: float,
    test_size: float,
    shuffle: bool = True,
    random_state: Optional[Union[int, np.random.RandomState]] = None,
) -> Tuple[np.array, Optional[np.array], np.array]:
    """Perform a grouped train/test split without stratification.

    .. seealso::

        * :py:meth:`mofdscribe.splitters.utils.stratified_train_test_partition`:
          performs an un-grouped stratified train/test split
        * :py:meth:`mofdscribe.splitters.utils.grouped_stratified_train_test_partition`:
          performs an grouped stratified train/test split

    Args:
        groups (np.typing.ArrayLike): Data used for grouping.
        train_size (float): Size of the training set as fraction.
        valid_size (float): Size of the validation set as fraction.
        test_size (float): Size of the test set as fraction.
        shuffle (bool): If True, perform a shuffled split. Defaults to True.
        random_state (Union[int, np.random.RandomState], optional):
            Random state for the suffler. Defaults to None.

    Returns:
        Tuple[np.array, Optional[np.array], np.array]: Train, validation, test indices.
    """
    train_indices = []
    valid_indices: Optional[List[Any]] = []
    test_indices = []

    unique_groups = np.unique(groups)

    train_size, valid_size, test_size = get_train_valid_test_sizes(
        len(unique_groups), train_size, valid_size, test_size
    )

    train_groups, test_groups = train_test_split(
        unique_groups,
        train_size=train_size + valid_size,
        test_size=test_size,
        shuffle=shuffle,
        random_state=random_state,
    )

    if valid_size > 0:
        train_groups, valid_groups = train_test_split(
            train_groups,
            train_size=train_size,
            shuffle=shuffle,
            random_state=random_state,
        )

    # now, get the original indices
    train_indices = np.where(np.isin(groups, train_groups))[0]
    if valid_size > 0:
        valid_indices = np.where(np.isin(groups, valid_groups))[0]
    else:
        valid_indices = None
    test_indices = np.where(np.isin(groups, test_groups))[0]

    return train_indices, valid_indices, test_indices


def quantile_binning(values: np.typing.ArrayLike, q: Sequence[float]) -> np.array:
    """Use :py:meth:`pandas.qcut` to bin the values based on quantiles."""
    values = pd.qcut(values, q, labels=np.arange(len(q) - 1)).astype(int)
    return values


def check_fraction(train_fraction: float, valid_fraction: float, test_fraction: float) -> None:
    """Check that the fractions are all between 0 and 1 and that they sum up to 1."""
    for name, fraction in [
        ("train fraction", train_fraction),
        ("valid fraction", valid_fraction),
        ("test fraction", test_fraction),
    ]:
        if not (fraction <= 1) & (fraction >= 0):
            raise ValueError(
                f"{name} is {fraction}. However, train/valid/test fractions must be between 0 and 1."
            )

    logger.debug(
        f"Using fractions: train: {train_fraction}, valid: {valid_fraction}, test: {test_fraction}"
    )
    if not (train_fraction + valid_fraction + test_fraction) == 1:
        raise ValueError("Train, valid, test fractions must sum to 1.")


def no_group_warn(groups: Optional[np.typing.ArrayLike]) -> None:
    """Raise warning if groups is None."""
    if groups is None:
        logger.warning(
            "You are not using a grouped split."
            " However, for retricular materials, grouping is typically a good idea to avoid data leakage."
        )


def downsample_splits(splits, sample_frac):
    downsampled = []
    for split in splits:
        downsampled.append(np.random.choice(split, int(len(split) * sample_frac)))

    return tuple(downsampled)


def sort_arrays_by_len(arrays, sort=True):
    if sort:
        arrays = [np.array(array) for array in arrays]
        arrays.sort(key=len, reverse=True)
        return tuple(arrays)
    else:
        return tuple(arrays)



from matminer.featurizers.structure.sites import SiteStatsFingerprint

from mofdscribe.featurizers.hostguest import HostGuestFeaturizer


def test_host_guest_featurizer(floating_structure):
    """Test the HostGuestFeaturizer."""
    featurizer = HostGuestFeaturizer(
        featurizer=SiteStatsFingerprint.from_preset("SOAP_formation_energy"),
        aggregations=("mean",),
    )
    featurizer.fit([floating_structure])
    features = featurizer.featurize(floating_structure)
    labels = featurizer._featurizer.feature_labels()
    assert len(features) == 2 * len(labels)
